from typing import Any, Optional, Tuple, Union
import numpy as np
from astropy import log
import pint
from pint_pal.utils import apply_cut_flag, apply_cut_select

class DMXParameter:
    """
    Convenience class for DMX parameters.
    """
    aliases = {'idx':'index', 'val':'dmx_val', 'err':'dmx_err', 'ep':'epoch',
               'r1':'low_mjd', 'r2':'high_mjd', 'f1':'low_freq',
               'f2':'high_freq', 'mask':'toa_mask'}
    def __init__(self) -> None:
        """
        """
        self.idx = 0  # index label [int]
        self.val = 0.0  # DMX value [cm**-3 pc]
        self.err = 0.0  # DMX uncertainty [cm**-3 pc]
        self.ep = 0.0  # epoch [day]
        self.r1 = 0.0 # left bin edge [day]
        self.r2 = 0.0 # right bin edge [day]
        self.f1 = 0.0 # lowest frequency [MHz]
        self.f2 = 0.0 # highest frequency [MHz]
        self.mask = []  # Boolean index array for selecting TOAs

    def __setattr__(self, name: str, value: Any) -> None:
        name = self.aliases.get(name, name)
        object.__setattr__(self, name, value)

    def __getattr__(self, name: str) -> Any:
        if name == 'aliases':
            raise AttributeError  # http://nedbatchelder.com/blog/201010/surprising_getattr_recursion.html
        name = self.aliases.get(name, name)
        return object.__getattribute__(self, name)

    def print_dmx(self, range_only: bool = False, fit_flag: bool = True, fortran: bool = False) -> None:
        """
        Print TEMPO-style DMX parameter.

        range_only=True will print only the bin edges [day].
        fit_flag=False will set the fit flag to 0 instead of 1.
        fortran=True will use 'D' instead of 'e'.
        """
        if range_only:
            DMX_str = (
                      f'DMXR1_{self.idx:04d}{self.r1:16.5f}\n'
                      f'DMXR2_{self.idx:04d}{self.r2:16.5f}\n'
                      )
            print(DMX_str)
        else:
            fit_flag = int(fit_flag)
            DMX_str = (
                      f'DMX_{self.idx:04d}{self.val:18.8e}{fit_flag:3}{self.err:20.8e}\n'
                      f'DMXEP_{self.idx:04d}{self.ep:16.5f}\n'
                      f'DMXR1_{self.idx:04d}{self.r1:16.5f}\n'
                      f'DMXR2_{self.idx:04d}{self.r2:16.5f}\n'
                      f'DMXF1_{self.idx:04d}{self.f1:16.3f}\n'
                      f'DMXF2_{self.idx:04d}{self.f2:16.3f}\n'
                      )
            if fortran: DMX_str = DMX_str.replace('e','D')
            print(DMX_str)


def group_dates(toas: pint.toa.TOAs, group_width: float = 0.1) -> list:
    """
    Returns MJDs of groups of TOAs no wider than a specified amount.

    NB: originally from dmx_fixer.py.

    toas is a PINT TOA object.
    group_width specifies the maximum width of the TOA group [day]; it defines
        and 'epoch'.
    """

    mjds = toas.get_mjds().value

    group_mjds = [0.0]
    igroup = 0
    group_count = 0
    low_mjd = 0.0
    for mjd in sorted(mjds):  # loop over sorted MJDs
        if low_mjd > 0 and (mjd - low_mjd) > group_width:
            group_mjds[igroup] /= group_count
            igroup += 1
            group_count = 0.0
            group_mjds.append(0.0)
        group_mjds[igroup] += mjd
        group_count += 1
        low_mjd = mjd
    group_mjds[igroup] /= group_count

    group_mjds = np.array(group_mjds)

    return group_mjds


def get_dmx_ranges(
    toas: pint.toa.TOAs,
    bin_width: float = 1.0, 
    pad: float = 0.0, 
    strict_inclusion: bool = True,
    check: bool = True
) -> list:
    """
    Returns a list of low and high MJDs defining DMX ranges, covering all TOAs.

    NB: returned DMX ranges are sorted according to MJD.

    This emulates the TEMPO binning algorithm in $TEMPO/src/arrtim.f, but not
        exactly.

    toas is a PINT TOA object.
    bin_width is the largest permissible DMX bin width [d].
    pad is the additional amount of time [d] added to the beginning and end of
        every DMX range.
    strict_inclusion=True if TOAs exactly on a bin edge are not in the bin for
        the implemented DMX model.
    check=True executes check_dmx_ranges() on the TOAs and the output DMX
        ranges.
    """

    # Order the MJDs
    mjds = toas.get_mjds().value
    isort = mjds.argsort()
    mjds = mjds[isort]

    # Initialize lists to be returned
    low_mjds, high_mjds = [], []  # low and high MJDs (bin edges), respectively

    # Now step through the MJDs, starting with the first (earliest)
    remaining_mjds = np.copy(mjds)
    while(len(remaining_mjds)):
        left_bin_edge = remaining_mjds.min() - pad
        if strict_inclusion: left_bin_edge -= 1e-11  # ~1 us
        right_bin_edge = left_bin_edge + bin_width + pad
        if strict_inclusion: right_bin_edge += 1e-11  # ~1 us
        in_bin = (left_bin_edge <= remaining_mjds) & \
                (remaining_mjds <= right_bin_edge)
        not_in_bin = np.logical_not(in_bin)  # currently not used
        right_bin_edge = remaining_mjds[in_bin].max() + pad # shrink bin
        if strict_inclusion: right_bin_edge += 1e-11  # ~1 us
        if np.any(in_bin):  # this should always be True
            low_mjds.append(left_bin_edge)
            high_mjds.append(right_bin_edge)
            not_accounted_for = remaining_mjds > right_bin_edge
        else:
            msg = "No TOAs in proposed DMX bin. Something is wrong."
            log.warning(msg)
        # Update remaining TOA MJDs
        remaining_mjds = remaining_mjds[not_accounted_for]

    dmx_ranges = list(zip(low_mjds,high_mjds))  # Should already be sorted

    # Check that all TOAs are in a bin and only one bin, and no empty ranges
    if check: check_dmx_ranges(toas, dmx_ranges)

    return dmx_ranges


def get_gasp_dmx_ranges(
    toas: pint.toa.TOAs,
    group_width: float = 0.1, 
    bin_width: float = 15.0,
    pad: float = 0.0,
    strict_inclusion: bool = True,
    check: bool = True
) -> list:
    """
    Return a list of DMX ranges that group GASP TOAs into bins.

    NB: Adopted from dmx_fixer.py.

    toas is a PINT TOA object.
    group_width [day] is a kwarg passed to group_dates().
    bin_width is the largest permissible DMX bin width [d].
    pad is the additional amount of time [d] added to the beginning and end of
        every DMX range.
    strict_inclusion=True if TOAs exactly on a bin edge are not in the bin for
        the implemented DMX model.
    check=True executes check_dmx_ranges() on the TOAs and the output DMX
        ranges.
    """

    # Check that the TOAs contain GASP data
    bes = np.array(toas.get_flag_value('be')[0])  # backend flag
    if 'GASP' not in set(bes):
        return []

    # Only consider GASP data outside the GUPPI range (PUPPI came later)
    end_gasp_era = min(toas[bes == 'GUPPI'].get_mjds().value)
    mjds = toas.get_mjds().value
    toas = toas[(mjds < end_gasp_era) & (bes == 'GASP')]
    mjds = mjds[(mjds < end_gasp_era) & (bes == 'GASP')]

    # Initialize lists to be returned
    low_mjds, high_mjds = [], []  # low and high MJDs (bin edges), respectively

    # Pair up high and low frequency GASP observations
    febes = np.array(toas.get_flag_value('f')[0])  # frontend_backend flag
    low_obs_mjds = group_dates(toas[febes == 'Rcvr_800_GASP'], group_width)
    high_obs_mjds = group_dates(toas[febes == 'Rcvr1_2_GASP'], group_width)
    iclosest_high = [abs(mjd-high_obs_mjds).argmin() for mjd in low_obs_mjds]
    iclosest_low = [abs(mjd-low_obs_mjds).argmin() for mjd in high_obs_mjds]
    for ihigh,ilow in enumerate(iclosest_low):
        if ihigh == iclosest_high[ilow]:  # found a pair
            delta_mjd = abs(high_obs_mjds[ihigh] - low_obs_mjds[ilow])
            if delta_mjd < bin_width:  # if the pair is close enough
                left_bin_edge = min(high_obs_mjds[ihigh] - group_width,
                        low_obs_mjds[ilow] - group_width) - pad
                if strict_inclusion: left_bin_edge -= 1e-11  # ~1 us
                right_bin_edge = max(high_obs_mjds[ihigh] + group_width,
                        low_obs_mjds[ilow] + group_width) + pad
                if strict_inclusion: right_bin_edge += 1e-11  # ~1 us
                low_mjds.append(left_bin_edge)
                high_mjds.append(right_bin_edge)

    dmx_ranges = list(zip(low_mjds,high_mjds))

    # It's possible some TOAs are missed in the above
    masks = [(left_bin_edge < mjds) & (mjds < right_bin_edge) for \
            left_bin_edge,right_bin_edge in dmx_ranges]
    ndmx_per_toa = np.array(masks).astype(int).sum(axis=0)
    inone = np.where(ndmx_per_toa == 0)[0]
    if len(inone):  # see if we can expand a range to accommodate
        dmx_ranges = expand_dmx_ranges(toas[inone], dmx_ranges,
                bin_width=bin_width, pad=pad,
                strict_inclusion=strict_inclusion, add_new_ranges=False,
                check=False)

    # Check that GASP TOAs are in a bin and only one bin, and no empty ranges
    if check: check_dmx_ranges(toas, dmx_ranges)

    return dmx_ranges


def expand_dmx_ranges(
    toas: pint.toa.TOAs,
    dmx_ranges: list,
    bin_width: float = 1.0,
    pad: float = 0.0,
    strict_inclusion: bool = True,
    add_new_ranges: bool = False,
    check: bool = True
) -> list:
    """
    Expands DMX ranges to accommodate new TOAs up to a maximum bin width.

    Returns a list of DMX ranges.

    NB: returned DMX ranges are sorted according to MJD.

    toas is a PINT TOA object.
    dmx_ranges is a list of (low_mjd, high_mjd) pairs defining the DMX ranges;
        see the output of get_dmx_ranges().
    bin_width is the largest permissible DMX bin width [d].
    pad is the additional amount of time [d] added to the beginning and end of
        every DMX range.
    strict_inclusion=True if TOAs exactly on a bin edge are not in the bin for
        the implemented DMX model.
    add_new_ranges=True will add new DMX ranges if the TOAs cannot be
        accomodated.
    check=True executes check_dmx_ranges() on the TOAs and the new DMX ranges.
    """

    dmx_ranges = sorted(dmx_ranges, key=lambda dmx_range: dmx_range[0])

    if not len(dmx_ranges):  # in case an empty list was passed
        if add_new_ranges:
            dmx_ranges += get_dmx_ranges(toas, bin_width=bin_width, pad=pad,
                    strict_inclusion=strict_inclusion, check=check)
        return dmx_ranges

    # Get the TOAs that don't have a DMX bin (inone)
    masks, ibad, iover, iempty, inone, imult = check_dmx_ranges(toas,
            dmx_ranges, full_return=True, quiet=True)

    iremain = []  # for the TOAs that are not accommodated by expansion

    dmx_ranges = np.array(dmx_ranges)  # tuples don't support item assignment

    for itoa in inone:
        expanded = False
        mjd = toas.get_mjds().value[itoa]
        inearest = np.argmin(abs(mjd - dmx_ranges.flatten()))
        irange = int(inearest / 2)  # index of the nearest dmx range
        left_bin_edge = dmx_ranges[irange][0]
        right_bin_edge = dmx_ranges[irange][1]
        if (left_bin_edge < mjd) & (mjd < right_bin_edge):  # in a bin
            continue  # previous iteration of loop could now cover this TOA
        width = right_bin_edge - left_bin_edge
        if not (inearest % 2):  # TOA is outside left bin-edge, has smaller MJD
            delta_mjd = left_bin_edge - mjd  # must be positive
            if width + delta_mjd < bin_width:  # can the range be widened?
                dmx_ranges[irange][0] -= delta_mjd + pad
                if strict_inclusion: dmx_ranges[irange][0] -= 1e-11  # ~1 us
                expanded = True
        else:  # TOA is outside right bin-edge, has larger MJD
            delta_mjd = mjd - right_bin_edge  # must be positive
            if width + delta_mjd < bin_width:  # can the range be widened?
                dmx_ranges[irange][1] += delta_mjd + pad
                if strict_inclusion: dmx_ranges[irange][1] += 1e-11  # ~1 us
                expanded = True
        if not expanded:  # TOA still not included
            iremain.append(itoa)

    dmx_ranges = list(map(tuple, dmx_ranges)) # return to list of tuples

    if add_new_ranges:
        dmx_ranges += get_dmx_ranges(toas[iremain], bin_width=bin_width,
                pad=pad, strict_inclusion=strict_inclusion, check=False)
    dmx_ranges = sorted(dmx_ranges, key=lambda dmx_range: dmx_range[0])

    # Check that all TOAs are in a bin and only one bin, and no empty ranges
    if check: check_dmx_ranges(toas, dmx_ranges)

    return dmx_ranges


def check_dmx_ranges(
    toas: pint.toa.TOAs,
    dmx_ranges: list,
    full_return: bool = False,
    quiet: bool = False
) -> Union[Tuple[list, list, list, list, list, list],None]:
    """
    Ensures all TOAs match only one DMX bin and all bins have at least one TOA.

    Also checks for improperly set and overlapping bins.

    NB: range boundaries are exclusive (strict_inclusion).

    Returns (if full_return)--
        masks: TOA masks for each range.
        ibad: Indices of bad/improper ranges.
        iover: Indices of ranges that overlap with one another.
        iempty: Indices of ranges with no TOAs assigned.
        inone: Indices of TOAs not assigned to any range.
        imult: Indices of TOAs assigned to multiple ranges.

    toas is a PINT TOA object.
    dmx_ranges is a list of (low_mjd, high_mjd) pairs defining the DMX ranges;
        see the output of get_dmx_ranges().
    full_return=True will return the masks, indices of problematic ranges, and
        indices of problematic TOAs.
    quiet=True turns off the logged info.
    """

    mjds = toas.get_mjds().value

    masks = [(left_bin_edge < mjds) & (mjds < right_bin_edge) for \
            left_bin_edge,right_bin_edge in dmx_ranges]

    ibad, iover, iempty, inone, imult = [], [], [], [], []

    # Check for bad bins
    for irange,dmx_range in enumerate(dmx_ranges):
        left_bin_edge, right_bin_edge = dmx_range[0], dmx_range[1]
        if left_bin_edge == right_bin_edge or left_bin_edge > right_bin_edge:
            ibad.append(irange)
            if not quiet:
                msg = f"DMX range with pythonic index {irange}, correponding to the DMX range {dmx_ranges[irange]}, is improper."
                log.info(msg)
    if len(ibad) and not quiet:
        msg = f"{len(ibad)} DMX ranges are improper."
        log.warning(msg)

    # Check for overlapping bins
    for irange,dmx_range in enumerate(dmx_ranges):
        left_bin_edge, right_bin_edge = dmx_range[0], dmx_range[1]
        other_ranges = \
                np.array(dmx_ranges)[np.where(np.arange(len(dmx_ranges)) \
                != irange)[0]]
        for mjd in other_ranges.flatten():
            if (left_bin_edge < mjd) & (mjd < right_bin_edge):  # finds overlap
                iover.append(irange)
                if not quiet:
                    msg = f"DMX range with pythonic index {irange}, correponding to the DMX range {dmx_ranges[irange]}, overlaps with other ranges."
                    log.info(msg)
                break  # only find first overlap
    if len(iover) and not quiet:
        msg = f"{len(iover)} DMX ranges are overlapping with one another."
        log.warning(msg)

    # Check for empty bins
    ntoa_per_dmx_bin = np.array(masks).astype(int).sum(axis=1)
    if not np.all(ntoa_per_dmx_bin):
        iempty = np.where(ntoa_per_dmx_bin == 0)[0]
        for irange in iempty:
            if not quiet:
                msg = f"DMX range with pythonic index {irange}, correponding to the DMX range {dmx_ranges[irange]}, overlaps no TOAs."
                log.info(msg)
        if not quiet:
            msg = f"{len(iempty)} DMX ranges have no TOAs."
            log.warning(msg)

    # Check each TOA in exactly one bin
    ndmx_per_toa = np.array(masks).astype(int).sum(axis=0)
    if not np.all(ndmx_per_toa == 1):
        inone = np.where(ndmx_per_toa < 1)[0]  # not one bin
        for itoa in inone:
            if not quiet:
                msg = f"TOA with index {itoa} (MJD {mjds[itoa]}, {toas.get_freqs().value[itoa]} MHz) does not have a DMX range."
                log.info(msg)
        imult = np.where(ndmx_per_toa > 1)[0]  # multiple bins
        for itoa in imult:
            irange = list(np.where(np.array(masks).astype(int)[:,itoa] \
                    == 1)[0])
            if not quiet:
                msg = f"TOA with index {itoa} (MJD {mjds[itoa]}, {toas.get_freqs().value[itoa]} MHz) is in {ndmx_per_toa[itoa]} DMX ranges (with pythonic indices {irange})."
                log.info(msg)
        if not quiet:
            msg = f"{len(inone)} TOAs have no DMX range and {len(imult)} TOAs are in multiple DMX ranges."
            log.warning(msg)

    if full_return:
        return masks, ibad, iover, iempty, inone, imult


def get_dmx_mask(toas: pint.toa.TOAs, low_mjd: float, high_mjd: float, strict_inclusion: bool = True) -> np.ndarray:
    """
    Return a Boolean index array for selecting TOAs from toas in a DMX range.

    toas is a PINT TOA object of TOAs in the DMX bin.
    low_mjd is the left edge of the DMX bin.
    high_mjd is the right edge of the DMX bin.
    strict_inclusion=True if TOAs exactly on a bin edge are not in the bin for
        the implemented DMX model.
    """

    mjds = toas.get_mjds().value

    if strict_inclusion:
        mask = (low_mjd < mjds) & (mjds < high_mjd)
    else:  # NB: the following could allow for TOAs in multiple bins
        mask = (low_mjd <= mjds) & (mjds <= high_mjd)

    return mask


def get_dmx_epoch(toas: pint.toa.TOAs, weighted_average: bool = True) -> float:
    """
    Return the epoch of a DMX bin.

    toas is a PINT TOA object of TOAs in the DMX bin.
    weighted_average=True uses the TOA uncertainties as weights in determining
        the DMX epoch.
    """

    mjds = toas.get_mjds().value

    if weighted_average:
        errs = toas.get_errors().value  # us
        weights = errs**-2
    else:
        weights = np.ones(len(toas))

    epoch = np.average(mjds, weights=weights)

    return epoch


def get_dmx_freqs(toas: pint.toa.TOAs, mask: np.ndarray, allow_wideband: bool = True) -> Tuple[float, float]:
    """
    Return the lowest and highest frequency of the TOAs in a DMX bin.

    toas is a PINT TOA object containing all the relevant TOAs.
    mask is a boolean mask that identifies the TOAs in this DMX bin.
    allow_wideband=True will consider the -fratio and -bw flags in the
        determination of these frequencies, if toas contains wideband TOAs.
    """

    freqs = toas.get_freqs()[mask].value  # MHz
    high_freq = np.max(freqs, initial=0.)
    low_freq = np.min(freqs, initial=np.inf)

    if allow_wideband:
        # indices of wideband TOAs
        wb_mask = mask & (np.array(toas.get_flag_value('pp_dm')[0]) != None)
        # the following arrays will be empty if all TOAs are narrowband
        fratios = toas.get_flag_value('fratio')[0] # frequency ratio / WB TOA
        fratios = np.array(fratios)[wb_mask]
        bws = toas.get_flag_value('bw')[0]  # bandwidth [MHz] / WB TOA
        bws = np.array(bws)[wb_mask]
        low_freqs = bws.astype('float32') / (fratios.astype('float32') - 1)
        low_freq = min(low_freq, np.min(low_freqs, initial=np.inf))
        high_freqs = bws.astype('float32') + low_freqs
        high_freq = max(high_freq, np.max(high_freqs, initial=0.))

    return low_freq, high_freq


def check_frequency_ratio(
    toas: pint.toa.TOAs,
    dmx_ranges: list,
    frequency_ratio: float = 1.1,
    strict_inclusion: bool = True,
    allow_wideband: bool = True,
    invert: bool = False,
    quiet: bool = False
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Check that the TOAs in a DMX bin pass a frequency ratio criterion.

    Returns the indices of the TOAs and DMX ranges that pass the test.

    toas is a PINT TOA object.
    dmx_ranges is a list of (low_mjd, high_mjd) pairs defining the DMX ranges;
        see the output of get_dmx_ranges().
    frequency_ratio is the ratio of high-to-low frequencies in the DMX bin;
        the frequencies used are returned by get_dmx_freqs().
    strict_inclusion is a Boolean kwarg passed to get_dmx_mask().
    allow_wideband is a Boolean kwarg passed to get_dmx_freqs(); if True, the
        bandwidths of the wideband TOAs are considered in the calculation.
    invert=True will return the indices of the TOAs and DMX ranges that fail to
        pass the test.
    quiet=True turns off the logged info.
    """

    toa_mask = np.zeros(len(toas), dtype=bool)  # selects TOAs and ranges
    dmx_range_mask = np.zeros(len(dmx_ranges), dtype=bool)  # that pass

    nfail_toas = 0
    for irange,dmx_range in enumerate(dmx_ranges):
        low_mjd, high_mjd = dmx_range[0], dmx_range[1]
        mask = get_dmx_mask(toas, low_mjd, high_mjd,
                strict_inclusion=strict_inclusion)
        low_freq, high_freq = get_dmx_freqs(toas, mask,
            allow_wideband=allow_wideband)
        if high_freq / low_freq >= frequency_ratio:  # passes
            toa_mask += mask
            dmx_range_mask[irange] = True
        else:  # fails
            nfail_toas += np.sum(mask)
            if not quiet:
                msg = f"DMX range with pythonic index {irange}, correponding to the DMX range {dmx_ranges[irange]}, contains TOAs that do not pass the frequency ratio test (TOAs with MJDs {toas.get_mjds()[mask].value})."
                log.info(msg)

    nfail_ranges = sum(np.logical_not(dmx_range_mask))
    if not quiet and nfail_ranges:
        msg = f"{nfail_ranges} DMX ranges, which include {nfail_toas} TOAs, do not pass the frequency ratio test."
        log.warning(msg)

    if not invert:  #  return those that pass
        return np.arange(len(toas))[toa_mask], \
                np.arange(len(dmx_ranges))[dmx_range_mask]
    else:  # return those that fail
        return np.arange(len(toas))[np.logical_not(toa_mask)], \
                np.arange(len(dmx_ranges))[np.logical_not(dmx_range_mask)]


def check_solar_wind(
    toas: pint.toa.TOAs,
    dmx_ranges: list,
    model: pint.models.timing_model.TimingModel, 
    max_delta_t: float = 0.1,
    bin_width: float = 1.0,
    solar_n0: float = 5.0,
    allow_wideband: bool = True,
    strict_inclusion: bool = True,
    pad: float = 0.0,
    check: bool = True,
    return_only: bool = False,
    quiet: bool = False
) -> list:
    """
    Split DMX ranges based on influence of the solar wind.

    Returns a list of DMX ranges.

    Uses approximation for SW model until PINT's model is improved/working.

    NB: Adopted from dmx_fixer.py.

    toas is a PINT TOA object.
    model is a PINT model object, which is currently how one accesses the solar
        elogation angle of the TOAs.
    dmx_ranges is a list of (low_mjd, high_mjd) pairs defining the DMX ranges;
        see the output of get_dmx_ranges().
    max_delta_t is the time delay [us] above which a DMX range will be split.
    bin_width is the largest permissible DMX bin width [d] for the new ranges.
    solar_n0 is the solar wind electron desity [cm**-3] at 1 AU to use in the
        model.
    allow_wideband is a Boolean kwarg passed to get_dmx_freqs(); if True, the
        bandwidths of the wideband TOAs are considered in the calculation.
    strict_inclusion is a Boolean kwarg passed to get_dmx_mask and
        get_dmx_ranges().
    pad is a float kwarg passed to get_dmx_ranges().
    check=True executes check_dmx_ranges() on the TOAs and the output DMX
        ranges.
    return_only=True will return only the indices of problematic ranges and
        TOAs.
    quiet=True turns off the logged info.
    """
    # constants in the model
    #one_AU = 499.005  # 1 astronomical unit [s]
    one_AU = 4.8481e-6  # 1 astronomical unit [pc]
    Dconst = 2.41e-4  # "inverse" dispersion constant [MHz**-2 pc cm**-3 s**-1]

    # Get the solar elongation angle [rad]
    phis = model.sun_angle(toas, heliocenter=True, also_distance=False).value

    toa_mask = np.zeros(len(toas), dtype=bool)  # selects problem TOAs
    dmx_range_mask = np.zeros(len(dmx_ranges), dtype=bool)  # and ranges

    for irange,dmx_range in enumerate(dmx_ranges):
        low_mjd, high_mjd = dmx_range[0], dmx_range[1]
        mask = get_dmx_mask(toas, low_mjd, high_mjd,
                strict_inclusion=strict_inclusion)
        low_freq, high_freq = get_dmx_freqs(toas, mask,
            allow_wideband=allow_wideband)
        # Convert to time delay, using calc from David's code (fixed)
        theta = np.pi - phis[mask]  # rad
        #Excess DM from Solar wind (approximate)
        #delta_dm = theta * (solar_n0 / 10.) * 2.4098e-2 / \
        #        (one_AU * abs(np.sin(theta)))  # pc cm**-3
        delta_dm = solar_n0 * one_AU * theta / abs(np.sin(theta))
        dm_delays = delta_dm / (Dconst * low_freq**2) * 1e6  # us
        delta_t = max(dm_delays) - min(dm_delays)
        if delta_t > max_delta_t:
            toa_mask += mask
            dmx_range_mask[irange] = True
            if not quiet:
                msg = f"DMX range with pythonic index {irange}, correponding to the DMX range {dmx_ranges[irange]}, contains TOAs that are affected by the solar wind (TOAs with MJDs {toas.get_mjds()[mask].value})."
                log.info(msg)
    nsolar = sum(dmx_range_mask)
    if not quiet and nsolar:
        msg = f"{nsolar} DMX ranges are affected by the solar wind."
        log.warning(msg)

    if return_only:  # return indices of problem TOAs and ranges
        return np.arange(len(toas))[toa_mask], \
                np.arange(len(dmx_ranges))[dmx_range_mask]
    else:  # return augmented ranges
        if nsolar:  # if there
            # Select good ranges
            dmx_ranges = np.array(dmx_ranges)[np.logical_not(dmx_range_mask)]
            dmx_ranges = list(map(tuple, dmx_ranges)) # return to tuples
            dmx_ranges += get_dmx_ranges(toas[toa_mask], bin_width=bin_width,
                    pad=pad, strict_inclusion=strict_inclusion)  # add new ones
            dmx_ranges = sorted(dmx_ranges, key=lambda dmx_range: dmx_range[0])

            # Check that all TOAs are in a bin and only one bin, no empty range
            if check: check_dmx_ranges(toas, dmx_ranges)

        return dmx_ranges


def add_dmx(model: pint.models.timing_model.TimingModel, bin_width: float = 1.0) -> None:
    """
    Checks for DispersionDMX and ensures the bin width is the only parameter.

    model is a PINT model object.
    bin_width is the largest permissible DMX bin width [d].
    """

    if 'DispersionDMX' not in model.components.keys():
        from pint.models.timing_model import Component
        dmx_component = Component.component_types["DispersionDMX"]
        model.add_component(dmx_component())
        dmx = model.components["DispersionDMX"]
        dmx.remove_DMX_range(1)  # remove the range it automatically initiates
        dmx.DMX.set(bin_width)
    else:
        dmx = model.components["DispersionDMX"]
        dmx.DMX.set(bin_width)


def model_dmx_params(model: pint.models.timing_model.TimingModel) -> Tuple[list, np.ndarray, np.ndarray]:
    """
    Get DMX ranges, values, and uncertainties from a PINT model object.

    model is a PINT model object.
    """

    dmx_ranges, dmx_vals, dmx_errs = [], [], []

    if 'DispersionDMX' in model.components.keys():
        dmx = model.components['DispersionDMX']
        idxs = dmx.get_indices()
        for idx in idxs:
            low_mjd = getattr(model, f"DMXR1_{idx:04d}").value
            high_mjd = getattr(model, f"DMXR2_{idx:04d}").value
            dmx_val = getattr(model, f"DMX_{idx:04d}").value
            try: dmx_err = getattr(model, f"DMX_{idx:04d}").uncertainty.value
            except AttributeError: dmx_err = 0.0
            dmx_ranges.append((low_mjd, high_mjd))
            dmx_vals.append(dmx_val)
            dmx_errs.append(dmx_err)
        dmx_vals = np.array(dmx_vals)
        dmx_errs = np.array(dmx_errs)

    return dmx_ranges, dmx_vals, dmx_errs


def remove_all_dmx_ranges(model: pint.models.timing_model.TimingModel, quiet: bool = False) -> None:
    """
    Uses PINT to remove all DMX parameter ranges from a timing model.

    model is a PINT model object.
    quiet=True turns off the logged info.
    """

    if 'DispersionDMX' in model.components.keys():
        dmx = model.components['DispersionDMX']
        idxs = dmx.get_indices()
        for idx in idxs:
            dmx.remove_DMX_range(idx)  #remove parameters
        if not quiet:
            msg = f"Removed {len(idxs)} DMX parameters from timing model."
            log.info(msg)
    else:
        pass


def setup_dmx(
    model: pint.models.timing_model.TimingModel,
    toas: pint.toa.TOAs,
    quiet: bool = True,
    frequency_ratio: float = 1.1,
    max_delta_t: float = 0.1,
    bin_width: Optional[float] = None,
    freeze_DM: bool = True
) -> pint.toa.TOAs:
    """
    Sets up and checks a DMX model using a number of defaults.

    Returns new PINT TOA object containing TOAs passing frequency ratio test.

    model is a PINT model object.
    toas is a PINT TOA object.
    frequency_ratio is the ratio of high-to-low frequencies in the DMX bin;
        the frequencies used are returned by get_dmx_freqs().
    max_delta_t is the time delay [us] above which a DMX range will be split.
    quiet=True turns off some of the logged warnings and info.
    bin_width=constant bin width if provided, otherwise use observatory defaults if None
    freeze_DM=True ensures the mean DM parameter is not fit.
    """

    # Freeze DM
    model.DM.frozen = freeze_DM

    # Get existing DMX ranges and values from model; adjust the mean DM
    old_dmx_ranges, old_dmx_vals, old_dmx_errs = model_dmx_params(model)
    if len(old_dmx_ranges) and np.sum(old_dmx_errs):
        idmx = old_dmx_errs != 0.0
        mean_old_dmx_val, sum_of_weights = np.average(old_dmx_vals[idmx],
                weights=old_dmx_errs[idmx]**-2, returned=True)
        mean_old_dmx_val_err = sum_of_weights**-0.5
        if abs(mean_old_dmx_val / mean_old_dmx_val_err) > 1.0:
            adjust_old_dmx = True
            old_DM = model.DM.value
            DM = model.DM.value + mean_old_dmx_val
            model.DM.set(DM)
            msg = f"Updated mean DM parameter from {old_DM:.7f} to {DM:.7f}."
            log.info(msg)
        else:
            adjust_old_dmx = False
    else:
        adjust_old_dmx = False

    # Set up DMX model
    if bin_width is None: #use observatory defaults
        if toas.observatories == set(['arecibo']): 
            bin_width = 0.5  # day
        else:
            bin_width = 6.5  #day

    # Calculate GASP-era ranges, if applicable
    dmx_ranges = get_gasp_dmx_ranges(toas, group_width=0.1, bin_width=15.0,
            pad=0.05, check=False)
    # Now expand to include all TOAs
    dmx_ranges = expand_dmx_ranges(toas, dmx_ranges, bin_width=bin_width,
            pad=0.05, add_new_ranges=True, check=False)

    # Ensure DM events have fine DMX binning --> need to generalize later
    if model.PSR.value == 'J1713+0747':
        toa_mask = np.zeros(len(toas), dtype=bool)
        dmx_range_mask = np.ones(len(dmx_ranges), dtype=bool)
        for irange,dmx_range in enumerate(dmx_ranges):
            low_mjd, high_mjd = dmx_range[0], dmx_range[1]
            # if it overlaps with the event at all...
            if ((57508 < low_mjd) and (low_mjd < 57513)) or \
                    ((57508 < high_mjd) and (high_mjd < 57513)):
                        dmx_range_mask[irange] = False
                        toa_mask += get_dmx_mask(toas, low_mjd, high_mjd)
        # Select the non-event ranges
        dmx_ranges = np.array(dmx_ranges)[dmx_range_mask]
        dmx_ranges = list(map(tuple, dmx_ranges))
        # Add new ranges based on the TOAs during the event
        dmx_ranges += get_dmx_ranges(toas[toa_mask], bin_width=0.5,
                check=False)
        # Sort the ranges
        dmx_ranges = sorted(dmx_ranges, key=lambda dmx_range: dmx_range[0])

    # Do basic checks of DMX model
    dmx_ranges = check_solar_wind(toas, dmx_ranges, model,
            max_delta_t=max_delta_t, bin_width=0.5, pad=0.05, check=False,
            quiet=quiet)
    itoas, iranges = check_frequency_ratio(toas, dmx_ranges,
            frequency_ratio=frequency_ratio, quiet=quiet)

    # Find TOAs failing fratio test and apply cuts
    ftoas, franges = check_frequency_ratio(toas, dmx_ranges,
            frequency_ratio=frequency_ratio, quiet=quiet, invert=True)
    fratio_inds = toas.table['index'][ftoas]
    if len(fratio_inds):
        apply_cut_flag(toas,fratio_inds,'dmx')
        apply_cut_select(toas,reason='frequency ratio check')

    dmx_ranges = np.array(dmx_ranges)[iranges]
    dmx_ranges = list(map(tuple, dmx_ranges))

    # Sort the ranges
    dmx_ranges = sorted(dmx_ranges, key=lambda dmx_range: dmx_range[0])

    # Check for sanity and use inone to flag dmx cuts
    masks, ibad, iover, iempty, inone, imult = \
            check_dmx_ranges(toas, dmx_ranges, full_return=True, quiet=False)

    if len(ibad) + len(iover) + len(iempty) + len(inone) + len(imult) == 0:
        msg = "Proposed DMX model OK."
        log.info(msg)
    else:
        pass  # check_dmx_ranges will print lots of warnings if quiet=False

    if len(dmx_ranges) == len(old_dmx_ranges):
        old_dmx_ranges = sorted(old_dmx_ranges,  #  sort the old ranges
                key=lambda dmx_range: dmx_range[0])
        if np.isclose(np.array(dmx_ranges).flatten(), \
                np.array(old_dmx_ranges).flatten(), \
                rtol=1e-10, atol=1e-5).all():
            msg = f"Proposed DMX ranges are the same as input; keeping input DMX model."
            log.info(msg)
            if adjust_old_dmx:
                dmx = model.components['DispersionDMX']
                idxs = dmx.get_indices()
                for idx in idxs:
                    low_mjd = getattr(model, f"DMXR1_{idx:04d}").value
                    high_mjd = getattr(model, f"DMXR2_{idx:04d}").value
                    dmx_val = getattr(model, f"DMX_{idx:04d}").value
                    dmx_val -= mean_old_dmx_val
                    frozen = getattr(model, f"DMX_{idx:04d}").frozen
                    dmx.remove_DMX_range(idx)
                    dmx.add_DMX_range(low_mjd, high_mjd, idx, dmx_val,
                            frozen=frozen)
                    msg = f"Subtracted {mean_old_dmx_val:.7f} from existing DMX values."
                log.info(msg)
            return toas

    # Remove all DMX bins from model
    remove_all_dmx_ranges(model, quiet=False)

    # Add DMX parameters to model
    add_dmx(model, bin_width)
    dmx = model.components['DispersionDMX']
    for irange,dmx_range in enumerate(dmx_ranges):
        dmx.add_DMX_range(dmx_range[0], dmx_range[1], index=irange+1,
                dmx=0.0, frozen=False)
    msg = f"Added {len(dmx_ranges)} DMX parameters to timing model."
    log.info(msg)

    return toas


def make_dmx(
    toas: pint.toa.TOAs,
    dmx_ranges: list,
    dmx_vals: Optional[np.ndarray] = None,
    dmx_errs: Optional[np.ndarray] = None,
    strict_inclusion: bool = True,
    weighted_average: bool = True,
    allow_wideband: bool = True,
    start_idx: int = 1,
    print_dmx: bool = False
):
    """
    Uses convenience functions to assemble a TEMPO-style DMX parameters.

    Returns list of DMXParameter objects.

    toas is a PINT TOA object.
    dmx_ranges is a list of (low_mjd, high_mjd) pairs defining the DMX ranges;
        see the output of get_dmx_ranges().
    dmx_vals is an array of DMX parameter values [pc cm**-3]; defaults to
        zeros.
    dmx_errs is an array of DMX parameter uncertainties [pc cm**-3]; defaults
        to zeros.
    strict_inclusion is a Boolean kwarg passed to get_dmx_mask().
    weighted_average is a Boolean kwarg passed to get_dmx_epoch().
    allow_wideband is a Boolean kwarg passed to get_dmx_freqs().
    start_idx is the index label of the first DMX parameter, which will
        increment with each entry in dmx_ranges; there is no assumption about
        MJD ordering.
    print_dmx=True will print all DMX parameters in TEMPO style.
    """

    dmx_parameters = []

    if dmx_vals is None:
        dmx_vals = np.zeros(len(toas))
        dmx_errs = np.zeros(len(toas))

    for irange,idx in enumerate(range(start_idx, start_idx+len(dmx_ranges))):
        low_mjd = min(dmx_ranges[irange])
        high_mjd = max(dmx_ranges[irange])
        mask = get_dmx_mask(toas, low_mjd, high_mjd, strict_inclusion)
        epoch = get_dmx_epoch(toas[mask], weighted_average)
        low_freq, high_freq = get_dmx_freqs(toas, mask, allow_wideband)
        dmx_parameter = DMXParameter()
        dmx_parameter.idx = idx
        dmx_parameter.val = dmx_vals[irange]
        dmx_parameter.err = dmx_vals[irange]
        dmx_parameter.ep = epoch
        dmx_parameter.r1 = low_mjd
        dmx_parameter.r2 = high_mjd
        dmx_parameter.f1 = low_freq
        dmx_parameter.f2 = high_freq
        dmx_parameter.mask = mask
        dmx_parameters.append(dmx_parameter)
        if print_dmx: dmx_parameter.print_dmx(range_only=False, fit_flag=True,
                fortran=False)

    return dmx_parameters



import itertools
import logging
import os.path
import subprocess
import textwrap
import tempfile
from collections import defaultdict
from io import StringIO

import matplotlib.pyplot as plt
from IPython.display import Markdown, display


class Report:
    """A report that is gradually built up.
    
    The idea is that an object like this is created at the beginning of
    a notebook, and various cells report various things, generating plots
    and writing markdown/math output both to the report and the notebook
    cells. At the end the report can be rendered to PDF.
    
    The report is organized into "sections"; each section is appended to
    by successive cells, and the sections are emitted in the order they 
    were first encountered. The intention is that sections can have a
    short name to refer to them by and a long name that appears in the
    document.
    
    This could probably be expanded to include page headers and footers
    for the final PDF; PDF generation goes through pandoc, which is quite
    configurable (and HTML might allow for such things too).
    """

    def __init__(self, *, title="", sections=None, figure_location=None):
        self.title = title
        self.sections = [] if sections is None else sections
        self.section_titles = {}
        self.section_content = defaultdict(StringIO)
        if figure_location is None:
            self.temporary_directory_object = tempfile.TemporaryDirectory()
            # Will go away when (if) this report is garbage collected
            figure_location = self.temporary_directory_object.name
        self.figure_location = figure_location

    def _ensure_section(self, section):
        if section not in self.sections:
            self.sections.append(section)

    def new_section(self, section):
        self._ensure_section(section)
        self.section_content[section] = StringIO()

    def add_markdown(self, section, content, *, also_display=True):
        self._ensure_section(section)
        if also_display:
            display(Markdown(content))
        self.section_content[section].write("\n\n")
        self.section_content[section].write(content)

    def add_verbatim(
        self, section, content, *, highlight_language="", also_display=True
    ):
        new_content = f"```{highlight_language}\n{content}\n```\n"
        self.add_markdown(section, new_content, also_display=also_display)

    # I'm not sure about tables - Markdown tables are not too annoying
    # and it's not super clear what format users would pass us tables in
    # (Astropy Tables? lists of lists they'd have to make themselves?)
    # https://www.markdownguide.org/extended-syntax/#tables

    def _new_figure_filename(self, section):
        for i in itertools.count(start=1):
            name = os.path.join(self.figure_location, f"figure-{section}-{i:04d}.png")
            if not os.path.exists(name):
                return name

    def add_plot(
        self,
        section,
        figure=None,
        *,
        dpi=300,
        width=1.0,
        caption="",
        also_display=False,
        **savefig_kwargs,
    ):
        self._ensure_section(section)
        # Include caption?
        filename = self._new_figure_filename(section)
        if figure is None:
            figure = plt.gcf()
        figure.savefig(filename, dpi=dpi, **savefig_kwargs)
        new_content = f"![{caption}]({filename}){{width={100*width:g}%}}\n\n"
        self.add_markdown(section, new_content, also_display=also_display)

    def generate(self, *, include_title=True):
        with StringIO() as o:
            if self.title and include_title:
                print(f"# {self.title}\n\n", file=o)
            for s in self.sections:
                t = self.section_titles.get(s, s)
                print(f"\n\n## {t}\n\n", file=o)
                print(self.section_content[s].getvalue(), file=o)
            return o.getvalue()

    def generate_pdf(self, pdf_filename):
        subprocess.run(
            [
                "pandoc",
                "--from",
                "markdown",
                "--to",
                "pdf",
                "--metadata",
                f"title={self.title}",
                "-o",
                pdf_filename,
                "--pdf-engine",
                "weasyprint",
            ],
            text=True,
            input=self.header + self.generate(include_title=False),
        )

    # We will display bold in red
    # The page margins are supplemented by the margins in the web HTML
    header = textwrap.dedent(
        """
        ---
        header-includes: |
          <style>
          body {
            max-width: 70em;
            background-color: #f0f0ff;
          }
          @media print {
            body {
                background-color: transparent;
                font-size: 10pt;
            }
          }
          @page {
            size: letter;
            margin: 1cm;
          }
          strong {
            color: #ff0000;
          }
          </style>
        ---
        """
    )

    def generate_html(self, html_filename):
        # The HTML is constructed from a template that can be viewed with `pandoc -D html`
        subprocess.run(
            [
                "pandoc",
                "--from",
                "markdown",
                "--to",
                "html",
                "-s",  # standalone, i.e., complete HTML including <html>
                "--self-contained",  # include images inline
                "--mathjax",
                "--metadata",
                f"title={self.title}",
                "-o",
                html_filename,
            ],
            text=True,
            input=self.header + self.generate(include_title=False),
        )

    def begin_capturing_log(self, section, introduction="", *, level=logging.WARNING):
        self._ensure_section(section)
        print(introduction +"\n", file=self.section_content[section])
        # This call ensures that logging is fully initialized before we add things
        # Without this the notebook loses the log messages
        logging.debug("Starting log capturing to report")
        report_log = logging.StreamHandler(self.section_content[section])
        report_log.setLevel(level)
        report_log.setFormatter(
            logging.Formatter("- `%(name)s`: %(levelname)s - %(message)s")
        )
        logging.getLogger().addHandler(report_log)



""" This is a set of utilities which check par files for completeness """

import re
import copy
from astropy import log
import astropy.units as u
import pint_pal.config
from pint.modelutils import model_equatorial_to_ecliptic

def check_if_fit(model, *param):
    """
    Check if a parameter or set of parameters
    exists and are fit

    Parameters
    ==========
    model: PINT model object
    param: parameter string(s)

    Raises
    ======
    ValueError
        If (any) parameter is frozen,
        the parameter is not present
    """
    for p in param:
        if not hasattr(model, p):
            raise ValueError("%s parameter is not present in par file."%p)
        if getattr(model, p).frozen:
            raise ValueError("%s parameter is frozen."%p)
    return


def has_and_check_if_fit(model, *param):
    """
    Convenience function for
    if hasattr(model, param):
        check_if_fit(model, param)

    Similar to check_if_fit() except that it will first
    test to see if the parameter exists, and if it does not,
    no error will be issued

    Parameters
    ==========
    model: PINT model object
    param: parameter string(s)
    """
    for p in param:
        if hasattr(model, p):
            check_if_fit(model, p)



def check_name(model):
    """Check the pulsar name

    Parameters
    ==========
    model: PINT model object

    Warnings
    ========
    UserWarning
        If PSR parameter is not in standard J epoch format, issue a warning.
    """
    name = model.PSR.value
    if not re.match("^((J[0-9]{4}[+-][0-9]{4})|(B[0-9]{4}[+-][0-9]{2}))$", name):
        msg = "PSR parameter is not in the proper format."
        log.warning(msg)
    else:
        msg = f"PSR parameter is {name}."
        log.info(msg)
    return


def check_spin(model):
    """Check spin parameters in par file.

    Parameters
    ==========
    model: PINT model object

    Raises
    ======
    ValueError
        If parameters provided are fixed
    """

    check_if_fit(model, "F0", "F1")

    # check for spindown parameters other than F0 or F1, i.e., any
    # such parameter that consists of the letter 'F' followed by
    # an integer
    for p in model.components['Spindown'].params:
        if p.startswith('F') and len(p)>1 and p[1:].isdigit():
            n = int(p[1:])
            if n < 0 or n >1:
                msg = 'Unexpected spin parameter %s should be removed' % (p,)
                log.warning(msg)

    return


def check_astrometry(model):
    """Check astrometry parameters in par file.

    Parameters
    ==========
    model: PINT model object

    Raises
    ======
    ValueError
        If parameters are not in ecliptic coordinates, or
        any astrometric parameter is fixed.
    """
    if "AstrometryEcliptic" not in model.components.keys():
        raise ValueError("Astrometric parameters not in ecliptic coordinates.")
    check_if_fit(model, "ELONG", "ELAT", "PMELONG", "PMELAT", "PX")
    return


def check_binary(model):
    """
    Check binary parameters in par file.
    The following models are checked for: ELL1, ELL1H, DD, DDK
    Performs multiple consistency checks for parameters and subsequent parameters,
    making sure they are fit.

    Parameters
    ==========
    model: PINT model object

    Raises
    ======
    ValueError
        If parameters are frozen or models are improperly defined
    """
    if not hasattr(model, 'binary_model_name'):
        return

    name = getattr(model, 'binary_model_name')

    if name == "ELL1":
        # Check base models, including either PB or FB formulations
        check_if_fit(model, "A1", "TASC", "EPS1", "EPS2")
        if hasattr(model, "PB") and hasattr(model, "FB0"):
            raise ValueError("Both PB and FB0 are defined")
        elif hasattr(model, "PB"):
            formalismFB = False
            check_if_fit(model, "PB")
        elif hasattr(model, "FB0"):
            formalismFB = True
            check_if_fit(model, "FB0")
        else:
            raise ValueError("Either PB or FB0 are required in the ELL1 model.")

        ### Need to include further checks here for higher-order FB terms,
        ### and to make sure they are sequentially fit for


        if not formalismFB:
            has_and_check_if_fit(model, "PBDOT")
            if hasattr(model, "SINI") and hasattr(model, "M2"):
                check_if_fit(model, "SINI", "M2")
            elif hasattr(model, "SINI") or hasattr(model, "M2"):
                raise ValueError("Both SINI and M2 are required when fitting for companion mass.")

        has_and_check_if_fit(model, "XDOT")
        if hasattr(model, "EPS1DOT") and hasattr(model, "EPS2DOT"):
            check_if_fit(model, "EPS1DOT", "EPS2DOT")
        elif hasattr(model, "EPS1DOT") or hasattr(model, "EPS2DOT"):
            raise ValueError("Both EPS1DOT and EPS2DOT are required when fitting for eccentricity derivatives.")



    elif name == "ELL1H":
        check_if_fit(model, "A1", "PB", "TASC", "EPS1", "EPS2", "H3") #H3 is rquired for ELL1H?
        has_and_check_if_fit(model, "PBDOT", "XDOT", "H4")

        if hasattr(model, "EPS1DOT") and hasattr(model, "EPS2DOT"):
            check_if_fit(model, "EPS1DOT", "EPS2DOT")
        elif hasattr(model, "EPS1DOT") or hasattr(model, "EPS2DOT"):
            raise ValueError("Both EPS1DOT and EPS2DOT are required when fitting for eccentricity derivatives.")



    elif name == "DD":
        check_if_fit(model, "A1" "E", "T0", "PB", "OM")
        has_and_check_if_fit(model, "PBDOT", "XDOT", "OMDOT", "EDOT")

        if hasattr(model, "SINI") and hasattr(model, "M2"):
            check_if_fit(model, "SINI", "M2")
        elif hasattr(model, "SINI") or hasattr(model, "M2"):
            raise ValueError("Both SINI and M2 are required when fitting for companion mass.")
    elif name == "DDK":
        check_if_fit(model, "A1", "E", "T0", "PB", "OM", "M2", "K96", "KOM", "KIN")
        has_and_check_if_fit(model, "PBDOT", "XDOT", "OMDOT", "EDOT")

        if hasattr(model, "SINI") and hasattr(model, "M2"):
            check_if_fit(model, "SINI", "M2")
        #elif hasattr(model, "SINI") or hasattr(model, "M2"):
    elif name == "T2":

        ### Are there plans for PINT to support the T2 model?

        check_if_fit(model, "A1", "ECC", "T0", "PB", "OM", "M2", "KOM", "KIN")
        # maybe check if SINI = KIN?
    else:
        raise ValueError("Atypical binary model used")

def check_jumps(model,receivers,toa_type=None):
    """Checks for correct type/number of JUMP/DMJUMPs in the par file

    Parameters
    ==========
    model: PINT model object
    receivers: list of receiver strings

    Raises
    ======
    ValueError
        If there is not one un-jumped receiver at the end, or
        no TOAs were recorded with that receiver

    ValueError
        If not all receivers are dm-jumped at the end, or
        no TOAs were recorded with that receiver
    """
    jumps = []
    rcvrs = copy.copy(receivers)

    for p in model.params:
        if p.startswith("JUMP"):
            jumps.append(p)

    for jump in jumps:
        j = getattr(model, jump)
        value = j.key_value[0]
        if value not in rcvrs:
            raise ValueError("Receiver %s not used in TOAs"%value)
        rcvrs.remove(value)

    length = len(rcvrs)
    if length > 1:
        raise ValueError("%i receivers require JUMPs"%length)
    elif length == 0:
        raise ValueError("All receivers have JUMPs, one must be removed")

    # Check DMJUMPS for wideband models
    if toa_type == 'WB':
        dmjumps = []
        rcvrs = copy.copy(receivers)

        for p in model.params:
            if p.startswith("DMJUMP"):
                dmjumps.append(p)

        for dmjump in dmjumps:
            j = getattr(model, dmjump)
            value = j.key_value[0]
            if value not in rcvrs:
                raise ValueError("Receiver %s not used in TOAs"%value)
            rcvrs.remove(value)

        length = len(rcvrs)
        if length:
            raise ValueError("%i receivers require DMJUMPs"%length)
    else:
        pass

    return

def check_ephem(toa):
    """Check that the ephemeris matches the latest version.

    Parameters
    ==========
    model: PINT toa object

    Warnings
    ========
    UserWarning
        If ephemeris is not set to the latest version.
    """
    if toa.ephem != pint_pal.config.LATEST_EPHEM:
        msg = (f"Wrong Solar System ephemeris in use ({toa.ephem});"
               f" should be {pint_pal.config.LATEST_EPHEM}.")
        log.warning(msg)
    else:
        msg = f"Current Solar System ephemeris in use is {toa.ephem}."
        log.info(msg)
    return

def check_bipm(toa):
    """Check that BIPM correction matches the latest version.

    Parameters
    ==========
    model: PINT toa object

    Warnings
    ========
    UserWarning
        If BIPM correction is not set to the latest version.
    """
    if toa.clock_corr_info['bipm_version'] != pint_pal.config.LATEST_BIPM:
        msg = (f"Wrong bipm_version ({toa.clock_corr_info['bipm_version']});"
               f" should be {pint_pal.config.LATEST_BIPM}.")
        log.warning(msg)
    else:
        msg = f"BIPM version in use is {toa.clock_corr_info['bipm_version']}."
        log.info(msg)
    return

def check_ecliptic(model):
    """Check that the parfile uses ecliptic coordinates.

    Parameters
    ==========
    model: PINT toa object

    Warnings
    ========
    UserWarning
        If not all model components are in ecliptic coordinates.
    """
    # Convert to/add AstrometryEcliptic component model if necessary.
    if 'AstrometryEquatorial' in model.components:
        msg = "AstrometryEquatorial in model components; switching to AstrometryEcliptic."
        log.warning(msg)
        model_equatorial_to_ecliptic(model)
    elif 'AstrometryEcliptic' in model.components:
        msg = "AstrometryEcliptic in model components."
        log.info(msg)
    else:
        msg = "Neither AstrometryEcliptic nor AstrometryEquatorial in model components."
        log.warning(msg)
    return

def check_troposphere(model):
    """Check that the model will (or will not) correct for the troposphere.

    Parameters
    =========
    model: PINT toa object

    Warnings
    ========
    UserWarning
        If CORRECT_TROPOSPHERE is not set to the default boolean and/or if the component is not in the model.
    """
    if 'TroposphereDelay' not in model.components.keys():
        from pint.models.timing_model import Component
        troposphere_delay = Component.component_types["TroposphereDelay"]
        model.add_component(troposphere_delay())
        msg = "Added TroposphereDelay to model components."
        log.warning(msg)
    tropo = model.components['TroposphereDelay'].CORRECT_TROPOSPHERE.value
    if tropo != pint_pal.config.CORRECT_TROPOSPHERE:
        model.components['TroposphereDelay'].CORRECT_TROPOSPHERE.set(
            pint_pal.config.CORRECT_TROPOSPHERE
        )
        msg = "Switching CORRECT_TROPOSPHERE setting."
        log.warning(msg)
    tropo = model.components['TroposphereDelay'].CORRECT_TROPOSPHERE.value
    msg = f"CORRECT_TROPOSPHERE is set to {tropo}."
    log.info(msg)
    return

def check_planet_shapiro(model):
    """Check that the model will (or will not) correct for the planets' Shapiro delays.

    Parameters
    =========
    model: PINT toa object

    Warnings
    ========
    UserWarning
        If PLANET_SHAPIRO is not set to the default boolean and/or if the component is not in the model.
    """
    if 'SolarSystemShapiro' not in model.components.keys():
        from pint.models.timing_model import Component
        sss_delay = Component.component_types["SolarSystemShapiro"]
        model.add_component(sss_delay())
        msg = "Added SolarSystemShapiro to model components."
        log.warning(msg)
    sss = model.components['SolarSystemShapiro'].PLANET_SHAPIRO.value
    if sss != pint_pal.config.PLANET_SHAPIRO:
        model.components['SolarSystemShapiro'].PLANET_SHAPIRO.set(
            pint_pal.config.PLANET_SHAPIRO
        )
        msg = "Switching PLANET_SHAPIRO setting."
        log.warning(msg)
    sss = model.components['SolarSystemShapiro'].PLANET_SHAPIRO.value
    msg = f"PLANET_SHAPIRO is set to {sss}."
    log.info(msg)
    return

def check_settings(model, toas, check_these=['name', 'ephem', 'bipm',
    'ecliptic', 'troposphere', 'planet_shapiro', 'bad_lo_range','toa_release']):
    """Umbrella function to run numerous check_ functions."""
    requires_model = {'name', 'ecliptic', 'troposphere', 'planet_shapiro'}
    requires_toas = {'ephem', 'bipm', 'bad_lo_range', 'toa_release'}
    requires_both = {}
    for thing in check_these:
        if thing in requires_model:
            globals()[f"check_{thing}"](model)
        elif thing in requires_toas:
            globals()[f"check_{thing}"](toas)
        elif thing in requires_both:
            globals()[f"check_{thing}"](model,toas)
        else:
            log.warning(f"check_{thing} not executed.")
    return

def check_bad_lo_range(toas):
    """Check: no Arecibo TOAs exist in MJD range affected by bad LO (57984-58447)

    15-yr specific mitigation strategy for excising affected data. Will raise
    log.warning if Arecibo TOAs exist in this range.

    Parameters
    ==========
    toas: `pint.toa.TOAs` object
    """
    bad_lo_start = (toas.get_mjds() > 57984.0*u.d)
    bad_lo_end = (toas.get_mjds() < 58447.0*u.d)
    lo_check = (bad_lo_start & bad_lo_end)
    ao_check = (toas.get_obss() == 'arecibo')

    if any(ao_check & lo_check):
        log.warning('Data affected by Arecibo bad LO should be excised (add bad-range to .yaml).')
        print('''
Add the following line(s) to the ignore block of your .yaml file...

  bad-range:
  - [57984,58447,"PUPPI"]
        ''')
    elif any(ao_check):
        log.info('Arecibo data affected by bad LO (MJD 57984-58447) has been ignored.')

def check_toa_release(toas):
    """Check: TOAs being used are from the most recent release

    Parameters
    ==========
    toas: `pint.toa.TOAs` object
    """
    release_flags = toas.get_flag_value('ver')[0]
    if len(set(release_flags)) > 1:
        log.error(f'TOAs from multiple releases should not be combined: {set(release_flags)}')
    else:
        if release_flags[0] == pint_pal.config.LATEST_TOA_RELEASE:
            log.info(f'All TOAs are from the latest release ({pint_pal.config.LATEST_TOA_RELEASE}).')
        else:
            log.warning(f'TOAs in use are from an old release {release_flags[0]}, '
                        f'not {pint_pal.config.LATEST_TOA_RELEASE}; '
                        f'update tim-directory in the .yaml accordingly.')



"""
This module provides tools for checking the finality of yamls
and gathering results for hand-off to DWG
"""

from pint_pal.yamlio import *
from pint_pal.timingconfiguration import TimingConfiguration
from astropy import log
from datetime import datetime
import subprocess
import shutil
import argparse
import glob
import os
import numpy as np
from matplotlib import pyplot as plt
import pint_pal.lite_utils as lu
import pint.utils
import astropy.units as u

# accessible to functions here, apparently
PAL_PATH = "/home/jovyan/work/pint_pal/" # assume running from here?
INTERMED_PATH = "/nanograv/share/15yr/timing/intermediate/"
PAL_RESULTS = os.path.join(PAL_PATH,"results")
PAL_CONFIGS = os.path.join(PAL_PATH,"configs")

def make_release_dir(type, overwrite=False):
    """
    Make new release directory to contain latest results

    Parameters
    ==========
    type: str
        narrowband (nb) or wideband (wb)
    overwrite: bool, optional
        overwrite existing files if release directory already exists (default: False)
    """
    now = datetime.now()
    Ymd = now.strftime("%Y%m%d")

    cmd = subprocess.Popen(["git","rev-parse","--short","HEAD"], stdout=subprocess.PIPE, stderr=subprocess.PIPE) 
    githash, stderr = cmd.communicate()
    githash = githash.strip().decode() # decode() since a bytestring is returned initially (b'[hash]')
    release_dir = f"{INTERMED_PATH}{Ymd}.Release.{type}.{githash}"
    if not os.path.isdir(release_dir):
        log.info(f"Making new release directory: {release_dir}")
        os.mkdir(release_dir)
    elif os.path.isdir(release_dir) and overwrite:
        log.warning(f"Overwriting files in release directory: {release_dir}")
    else:
        log.warning(f"Release directory already exists: {release_dir}")

    return release_dir

def check_cleared(type):
    """
    Check that all yamls of the specified type have been cleared

    Parameters
    ==========
    type: str
        narrowband (nb), wideband (wb), or both (nbwb)
    """
    if type == "nbwb":
        yamls = glob.glob(f"{PAL_CONFIGS}/*.yaml")
    else:
        yamls = glob.glob(f"{PAL_CONFIGS}/*.{type}.yaml")
    for y in yamls:
        tc = TimingConfiguration(y)
        if not tc.get_check_cleared():
            log.warning(f"{tc.get_source()} has not been cleared.")
            
    return yamls

def check_dupes_copy(results, release_dir, add_base=None):
    """
    Check for duplicate results (copy if no duplicates)

    Parameters
    ==========
    results: list
        list (should be one element) of results file(s) to copy
    release_dir: str
        path to release directory
    add_base: str, optional
        optional basename added to file being copied
    """
    if len(results) != 1:
        log.warning("Multiple/no matching results files found.")
        print(results)
    else:
        file2copy = os.path.basename(results[0])
        if add_base:
            file2copy = f"{add_base}.{file2copy}"

        dest_file = f"{release_dir}/{file2copy}"
        shutil.copyfile(results[0], dest_file)

def locate_copy_results(yamls,type,destination=None):
    """
    Get latest results from yamls, copy to release directory

    Parameters
    ==========
    yamls: list
        yamls to use for locating latest results
    type: str
        narrowband (nb) or wideband (wb)
    destination: str
        path to release directory
    """
    for y in yamls:
        tc = TimingConfiguration(y)
        source = tc.get_source()
        noise_dir = tc.get_noise_dir()
        latest_yaml = [y]
        latest_par = [f"{PAL_PATH}{tc.get_model_path()}"]
        latest_tim = glob.glob(f"{noise_dir}results/{source}_*.tim") # underscore to avoid duplicating split-tel results
        noise_chains = glob.glob(f"{noise_dir}{source}_{type}/chain_1.txt")
        noise_pars = glob.glob(f"{noise_dir}{source}_{type}/pars.txt")

        log.info(f"Locating/copying files for {source}...")
        check_dupes_copy(latest_tim, destination)
        check_dupes_copy(latest_par, destination)
        check_dupes_copy(latest_yaml, destination)
        check_dupes_copy(noise_chains, destination, add_base=f"{source}.{type}")
        check_dupes_copy(noise_pars, destination, add_base=f"{source}.{type}")

def main():

    parser = argparse.ArgumentParser(
        description="Make release directory; copy final nb/wb TA data products there",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-t",
        "--type",
        type=str,
        nargs=1,
        help="Release type: nb, wb, or both (nbwb)",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="Overwrite files in release directory",
    )
    args = parser.parse_args()

    if not args.type:
        print("Please provide a release type.")
    elif args.type[0] not in ['nb','wb','nbwb']:
        print(args.type)
        print("Unrecognized release type.")
    else:
        if args.type[0] != "nbwb":
            # make directory
            rel_dir = make_release_dir(args.type[0], overwrite=args.overwrite)

            # get yamls
            yamls = check_cleared(args.type[0])

            # locate results and copy them to release directory
            locate_copy_results(yamls,args.type[0],rel_dir)

        else: # nbwb
            rel_dir = make_release_dir(args.type[0], overwrite=args.overwrite)
            nb_yamls = check_cleared('nb')
            locate_copy_results(nb_yamls,'nb',rel_dir) # works for nb/wb separately
            wb_yamls = check_cleared('wb')
            locate_copy_results(wb_yamls,'wb',rel_dir)

if __name__ == "__main__":
    main()



from pint_pal.yamlio import read_yaml, write_yaml
from astropy import log
import os
import glob
import shutil
import git

def shuffle_pars(tc):
    """
    Copy the current timing-model to the archive directory and set it as compare-model;
    then copy the most recent par file to the par directory and set it as timing-model.
    Automatically perform a `git add` in each case.

    Parameters
    ----------
    tc: TimingConfiguration object derived from the YAML configuration file.
    """
    # Get file and repo locations
    par_dir = tc.par_directory
    archive_dir = os.path.join(par_dir, "archive")
    repo = git.Repo(par_dir, search_parent_directories=True)

    timing_model = tc.get_model_path()
    timing_model_basename = os.path.split(timing_model)[1]
    new_compare_model = os.path.join(archive_dir, timing_model_basename)

    # Get path to new par file
    # Assummes the newest parfile in the working directory is the correct one
    new_par = max(glob.iglob("*.par"), key=os.path.getctime)
    new_par_basename = os.path.split(new_par)[1]
    new_timing_model = os.path.join(par_dir, new_par_basename)

    # Make sure there are no uncommitted changes to the current timing-model
    head = repo.commit("HEAD")
    tracked_files = set(item.path for item in head.tree.traverse())
    diff = head.diff(None) # diff between HEAD and current working directory state

    if os.path.relpath(timing_model, start=repo.working_tree_dir) not in tracked_files:
        log.info(f"Par file to be archived ({timing_model}) is not tracked. "
                 f"Copying it to {archive_dir} and leaving the original.")
        shutil.copy(timing_model, new_compare_model)
        repo.index.add(os.path.abspath(new_compare_model))
    else:
        log.info(f"Moving {timing_model} to {archive_dir}")
        shutil.move(timing_model, new_compare_model, copy_function=shutil.copy)
        repo.index.add(os.path.abspath(new_compare_model))

    log.info(f"Copying {new_par} to {par_dir}")             
    shutil.copy(new_par, new_timing_model)
    repo.index.add(os.path.abspath(new_timing_model))
    
    # Update YAML configuration accordingly
    config_path = os.path.abspath(tc.filename)
    config = read_yaml(config_path)
    config['timing-model'] = os.path.relpath(new_timing_model, start=par_dir)
    config['compare-model'] = os.path.relpath(new_compare_model, start=par_dir)
    write_yaml(config, config_path)
    repo.index.add(config_path)
 



import os
import argparse
from timingnotebook import TimingNotebook


## Argument parser setup
parser = argparse.ArgumentParser(description="""\
    NANOGrav Notebook Writer: script to generate NANOGrav-specific
    timing notebooks with the TimingNotebook class; the output notebook
    filename will resemble the input config filename
    (e.g. process-J1910+1256.nb.ipynb), unless otherwise specified.
    """)
parser.add_argument("-c", "--config", default=None, \
                    type=str, help="configuration (.yaml) filename")
parser.add_argument("-f", "--filename", default=None, \
                    type=str, help="Output filename")
parser.add_argument("-t", "--timdir", default=None, \
                    type=str, help="Path to directory with tim file")
parser.add_argument("-p", "--pardir", default=None, \
                    type=str, help="Path to directory with par file")
parser.add_argument("-a", "--autorun", action="store_true", \
                    help="For batch runs; removes markdown/plotting/interactive cells")
parser.add_argument("--use_existing_noise",action="store_true", \
                    help="Use existing noise analysis results")
parser.add_argument("--run_noise",action="store_true", \
                    help="Run noise analysis")
parser.add_argument("-w", "--write_results",action="store_true", \
                    help="Write resulting par/tim/dmx files")
parser.add_argument("--prenoise_only",action="store_true", \
                    help="Lite notebook with setup/prenoise only")
args = parser.parse_args()


tn = TimingNotebook()


tn.add_setup(autorun=args.autorun)
tn.add_prenoise(filename=args.config,tim_directory=args.timdir,
                par_directory=args.pardir,write=args.write_results,
                autorun=args.autorun)

if not args.prenoise_only:
    tn.add_noise(use_existing=args.use_existing_noise,run_noise=args.run_noise,
                 write=args.write_results,autorun=args.autorun)
    tn.add_compare(autorun=args.autorun)
    tn.add_significance(autorun=args.autorun)
    tn.add_summary(autorun=args.autorun)
    tn.add_changelog(autorun=args.autorun)

# Determine output filename
if args.filename is not None:
    outfile = args.filename
elif args.config is not None:
    config_only = args.config.split('/')[-1]
    config_base = os.path.splitext(config_only)[0]
    outfile = f"process-{config_base}.ipynb"
else:
    outfile = "process.ipynb"

tn.write_out(filename=outfile)



"""
This code contains the TimingNotebook class, which is used to
generate a pipeline notebook. Each section of cells has been
modularized into different functions.

Very basic usage:
    from timingnotebook import TimingNotebook
    tn = TimingNotebook()
    tn.add_import_cell()
    tn.write_out("process.ipynb")
"""


import os
from base64 import encodestring
import codecs
import textwrap

import nbformat as nbf
from nbformat.v4.nbbase import (
    new_code_cell, new_markdown_cell, new_notebook,
    new_output, new_raw_cell
)


METADATA = {
  "kernelspec": {
    "name": "python3",
    "display_name": "Python 3",
    "language": "python"
  },
  "language_info": {
    "codemirror_mode": {
      "version": 3,
      "name": "ipython"
    },
    "file_extension": ".py",
    "pygments_lexer": "ipython3",
    "version": "3.7.6",
    "mimetype": "text/x-python",
    "name": "python",
    "nbconvert_exporter": "python"
  }
}


class TimingNotebook:
    """
    The class that contains the functionality to add cells,
    store that information, and write to an ipynb file.
    """
    def __init__(self, metadata=METADATA):
        """
        Initialization method

        Parameters
        ==========
        metadata (optional): .ipynb metadata, with the standard defined in METADATA
        """
        self.cells = list()
        self.metadata = metadata


    def add_cell(self, text, mode="code", skip=False):
        """
        A helper function to make adding new cells consistent.
        Can be used to make a generic cell as well.

        Parameters
        ==========
        text : block string
        mode (optional) : type of cell to add
        """
        if not skip:
            if mode == "code":
                new_cell = new_code_cell
            elif mode == "markdown":
                new_cell = new_markdown_cell
            elif mode == "raw":
                new_cell = new_raw_cell
            self.cells.append(new_cell(source=textwrap.dedent(text)))
        else:
            pass

    # Convenience functions
    add_code_cell = lambda self, text: self.add_cell(text=text, mode="code")
    add_markdown_cell = lambda self, text: self.add_cell(text=text, mode="markdown")
    add_raw_cell = lambda self, text: self.add_cell(text=text, mode="raw")

    # skip provides flexibility to skip cells for batch runs (autorun)
    add_code_cell_skip = lambda self, text, skip: self.add_cell(text=text, mode="code", skip=skip)
    add_markdown_cell_skip = lambda self, text, skip: self.add_cell(text=text, mode="markdown", skip=skip)


    def add_setup(self,autorun=False):
        """ Add setup helper info, import and log.setLevel cells """
        self.add_markdown_cell_skip('''\
        # \[set-up\], imports

        Reminder (if working on the notebook server): grab current copies of relevant software before doing anything! Make sure your copy of `pint_pal` is up to date and you're working on a development branch, e.g. `psr/J1234+5678/jks`. Then:
        ```
        > cd ~/work/pint_pal/
        > pip install git+git://github.com/nanograv/pint --upgrade --user
        > pip install git+https://github.com/nanograv/enterprise.git --upgrade --user
        > pip install git+https://github.com/nanograv/enterprise_extensions.git --upgrade --user
        > pip install --user -e .
        ```\
        ''',autorun)
        self.add_code_cell('''\
        from pint_pal.lite_utils import *
        from pint_pal.par_checker import *
        from pint_pal.plot_utils import *
        from pint_pal.utils import *
        from pint_pal.dmx_utils import *
        import pint_pal.noise_utils as nu
        from pint_pal.ftester import run_Ftests
        from pint_pal.timingconfiguration import TimingConfiguration
        import yaml
        from astropy import log
        import pint.fitter
        from pint.utils import dmxparse
        import os
        from astropy.visualization import quantity_support
        quantity_support()

        # notebook gives interactive plots but not until the kernel is done
        %matplotlib notebook
        # inline gives non-interactive plots right away
        #%matplotlib inline\
        ''')
        self.add_code_cell_skip('''\
        log.setLevel("INFO") # Set desired verbosity of log statements (DEBUG/INFO/WARNING/ERROR)
        git_config_info()\
        ''',autorun)

    def add_prenoise(self,filename="config.yaml",
                     tim_directory=None, par_directory=None,
                     write=False,autorun=False):
        """ Add cells that load yaml/par/tim and do pre-noise fits """
        self.add_markdown_cell_skip('''\
        # develop/update \[prenoise\] timing solution

        Load configuration (`.yaml`) file, get TOAs and timing model; if you're running from the root of the git distribution, simply edit the `.yaml` file name, otherwise include relevant paths to the `.yaml` file, and `.par`/`.tim` directories as kwargs (see commented example).\
        ''',autorun)

        if filename is not None:
            # Allow for None to be passed if these are not strings.
            if isinstance(filename, str):
                filename = f'"{filename}"'
            if isinstance(tim_directory, str):
                tim_directory = f'"{tim_directory}"'
            if isinstance(par_directory, str):
                par_directory = f'"{par_directory}"'

            self.add_code_cell(f'''\
            tc = TimingConfiguration({filename}, tim_directory={tim_directory}, par_directory={par_directory})

            using_wideband = tc.get_toa_type() == 'WB'
            mo,to = tc.get_model_and_toas()\
            ''')
        else:
            # If no config file is provided, generate template version of notebook
            # Does not yet handle tim_directory/par_directory, but these should not be set with no config.
            self.add_code_cell('''\
            tc = TimingConfiguration("configs/[filename].yaml")
            # tc = TimingConfiguration("[path/to/config/filename].yaml", par_directory="[path/to/results]", tim_directory="[/path/to/wherever/the/tim/files/are]")

            using_wideband = tc.get_fitter() == 'WidebandTOAFitter'
            mo,to = tc.get_model_and_toas()\
            ''')

        self.add_markdown_cell_skip('''\
        Run basic checks for pulsar name, solar system ephemeris, clock correction, ecliptic coordinates, tropospheric delays, planet Shapiro delays, and if applicable, removal of Arecibo data affected by bad LO. Check that TOAs being used are from the latest `toagen` release. Also check for the appropriate number of receiver JUMPs and DMJUMPs and fix them automatically if necessary.\
        ''',autorun)
        self.add_code_cell('''\
        check_settings(mo,to)

        receivers = get_receivers(to)
        add_feJumps(mo,receivers)
        if using_wideband:
            add_feDMJumps(mo,receivers)
        check_jumps(mo,receivers,toa_type=tc.get_toa_type())\
        ''')
        self.add_markdown_cell_skip('''\
        Compute pulse numbers; this ensures that parameter changes made in the model will not break phase connection.\
        ''',autorun)
        self.add_code_cell_skip('''\
        to.compute_pulse_numbers(mo)\
        ''',autorun)

        if not autorun:
            self.add_markdown_cell('''\
            Ensure DMX windows are calculated properly for the current set of TOAs, set non-binary epochs to the center of the data span and print a summary of TOAs included.\
            ''')
            self.add_code_cell('''\
            to = setup_dmx(mo,to,frequency_ratio=tc.get_fratio(),max_delta_t=tc.get_sw_delay())
            center_epochs(mo,to)
            to.print_summary()\
            ''')
            self.add_markdown_cell('''\
            Define the fitter object and inspect prefit residuals vs. MJD (also vs. orbital phase for binary MSPs).\
            ''')
            self.add_code_cell('''\
            fo = tc.construct_fitter(to,mo)
            plot_residuals_time(fo, restype='prefit')
            if mo.is_binary:
                plot_residuals_orb(fo, restype='prefit')
            if using_wideband:
                plot_dm_residuals(fo, restype='prefit')\
            ''')
            self.add_markdown_cell('''\
            Check that free parameters follow convention, do the fit, plot post-fit residuals, write a pre-noise par file and print a summary of fit results.\
            ''')
            self.add_code_cell('''\
            fo.model.free_params = tc.get_free_params(fo)
            check_fit(fo,skip_check=tc.skip_check)\
            ''')
            self.add_code_cell('''\
            fo.fit_toas()
            plot_residuals_time(fo, restype='postfit')
            if mo.is_binary:
                plot_residuals_orb(fo, restype='postfit')
            if using_wideband:
                plot_dm_residuals(fo, restype='postfit')

            fo.print_summary()
            
            chi2_decrease = fo.resids_init.chi2-fo.resids.chi2
            print(f"chi-squared decreased during fit by {chi2_decrease}")
            if hasattr(fo, "converged") and fo.converged:
                print("Fitter has converged")
            else:
                if abs(chi2_decrease)<0.01:
                    print("Fitter has probably converged")
                elif chi2_decrease<0:
                    log.warning("Fitter has increased chi2!")
                else:
                    log.warning("Fitter may not have converged")
            if chi2_decrease > 0.01:
                log.warning("Input par file is not fully fitted")\
            ''')
            self.add_code_cell(f'''\
            write_prenoise = {write}
            if write_prenoise:
                write_par(fo,toatype=tc.get_toa_type(),addext='_prenoise')\
            ''')
        else:
            self.add_code_cell('''\
            to = setup_dmx(mo,to,frequency_ratio=tc.get_fratio(),max_delta_t=tc.get_sw_delay())
            center_epochs(mo,to)
            fo = tc.construct_fitter(to,mo)
            fo.model.free_params = tc.get_free_params(fo)
            check_fit(fo,skip_check=tc.skip_check)
            fo.fit_toas()\
            ''')
            if write:
                self.add_code_cell('''\
                write_par(fo,toatype=tc.get_toa_type(),addext='_prenoise')\
                ''')

    def add_noise(self,run_noise=False,use_existing=False,autorun=False,write=False):
        """ Add a number of cells that will perform noise modeling via enterprise """
        self.add_markdown_cell_skip('''\
        # \[noise\] analysis, re-fit

        Noise analysis runs are required for the 15-yr v0.9 data set, using the latest available timing model and set of TOAs. Once a new `.yaml` file is merged, noise runs will get kicked off on Thorny Flats (HPCs) using the `timing-model` and `toas` indicated therein. For pulsars without existing pre-noise solutions, timers should use the [prenoise] section of this notebook to improve TOA excision, ensure residuals are flat, then submit a merge request with that solution, and wait for noise results to be made available before proceeding.

        We strongly discourage running noise analyses from on the notebook server, since doing so can take several hours (or days!) to complete and hogs lots of shared resources. Set `run_noise_analysis = False` unless you have a good reason to do otherwise.\
        ''',autorun)

        # Allow these to be toggled in writing the notebook, via argparse
        self.add_code_cell(f'''\
        run_noise_analysis = {run_noise}
        use_existing_noise_dir = {use_existing}\
        ''')
        self.add_markdown_cell_skip('''\
        If `run_noise_analysis = True`, perform noise modeling using enterprise and enterprise_extensions; this cell will take a long time to run. Status can be monitored once modeling is 1% complete. New noise parameters will be added to the timing model if there are existing results or `model_noise` is run. Redefine the fitter object (`fo`), now with noise parameters, and re-fit.\
        ''',autorun)
        self.add_code_cell_skip('''\
        if run_noise_analysis or use_existing_noise_dir:
            remove_noise(mo)
            nu.model_noise(mo, to, using_wideband = using_wideband, run_noise_analysis = run_noise_analysis)
            mo = nu.add_noise_to_model(mo, using_wideband = using_wideband, base_dir=tc.get_noise_dir())

            fo = tc.construct_fitter(to,mo)
            fo.model.free_params = tc.get_free_params(fo)

            fo.fit_toas()
            plot_residuals_time(fo, restype='postfit')
            if mo.is_binary:
                plot_residuals_orb(fo, restype='postfit')
            fo.print_summary()\
        ''',autorun)

        # Remove plotting with autorun
        if autorun:
            self.add_code_cell('''\
            nu.model_noise(mo, to, using_wideband = using_wideband, run_noise_analysis = run_noise_analysis)
            if run_noise_analysis or use_existing_noise_dir:
                mo = nu.add_noise_to_model(mo, using_wideband = using_wideband)
                fo = tc.construct_fitter(to,mo)
                fo.model.free_params = tc.get_free_params(fo)
                fo.fit_toas()\
            ''')

        self.add_markdown_cell_skip('''\
        Write resulting `.par` and `.tim` files; also run PINT implementation of dmxparse and save the values in a text file with `save = True` (requires DMX bins in model to run properly).\
        ''',autorun)

        # Allow write to be toggled
        self.add_code_cell(f'''\
        write_results = {write}
        if write_results:
            outpar = None  # None leads to default string value
            write_par(fo,toatype=tc.get_toa_type(),outfile=outpar)
            outtim = None  # None leads to default string value
            write_tim(fo,toatype=tc.get_toa_type(),outfile=outtim)

            dmx_dict = dmxparse(fo, save=True)\
        ''')
        
    def add_compare(self,autorun=False):
        """ Add cells to compare timing models """
        self.add_markdown_cell_skip('''\
        # \[compare\] to previous timing model

        Compare post-fit model to `compare-model` (or pre-fit model, if `compare-model` is not specified in the `.yaml` file). Use `?mo.compare` for more information about verbosity options.\
        ''',autorun)
        self.add_code_cell_skip('''\
        compare_models(fo,model_to_compare=tc.get_compare_model(),verbosity='check',nodmx=True,threshold_sigma=3.)\
        ''',autorun)

    def add_significance(self,autorun=False):
        """ Add cells that calculate resid stats, do F-tests """
        self.add_markdown_cell_skip('''\
        # check parameter \[significance\]

        Get information on the weighted (W)RMS residuals per backend. Set `epoch_avg = True` to get the (W)RMS of the epoch-averaged residuals (does not work for wideband analysis; the timing model must have `ECORR` in order for epoch averaging to work). Set `whitened = True` to get the (W)RMS of the whitened residuals. Set both to `True` to get the (W)RMS of the whitened, epoch-averaged residuals.

        For wideband analysis, set `dm_stats = True` to also return the (W)RMS of the DM residuals.\
        ''',autorun)
        self.add_code_cell('''\
        if not using_wideband:
            rs_dict = resid_stats(fo, epoch_avg = True, whitened = True, print_pretty = True)
        else:
            rs_dict, dm_dict = resid_stats(fo, whitened = True, dm_stats = True, print_pretty = True)\
        ''')
        self.add_markdown_cell_skip('''\
        Run F-tests to check significance of existing/new parameters; `alpha` is the p-value threshold for rejecting the null hypothesis that a parameter is not significant. This cell may take 5-10 minutes to run.\
        ''',autorun)
        self.add_code_cell('''\
        savedLevel = log.getEffectiveLevel()
        try:
            log.setLevel("WARNING")
            Ftest_dict = run_Ftests(fo, alpha = 0.0027)
        finally:
            log.setLevel(savedLevel)\
        ''')

    def add_summary(self,autorun=False):
        """ Add cells that will generate summary pdfs """
        self.add_markdown_cell_skip('''\
        # generate \[summary\] pdf
        
        Generate summary plots required for pdf summaries. Note: this cell will output white space for the plots, but will save them and incorporate them into the pdf summaries appropriately.\
        ''',autorun)
        self.add_code_cell('''\
        if not using_wideband:
            plots_for_summary_pdf_nb(fo)
        else:
            plots_for_summary_pdf_wb(fo)\
        ''')
        self.add_code_cell('''\
        PARFILE = os.path.join(tc.par_directory,tc.config["timing-model"])
        if not using_wideband:
            pdf_writer(fo, PARFILE, rs_dict, Ftest_dict, append=None)
        else:
            pdf_writer(fo, PARFILE, rs_dict, Ftest_dict, dm_dict = dm_dict)\
        ''')

    def add_changelog(self,autorun=False):
        """ Add cell explaining how to generate changelog entries """
        self.add_markdown_cell_skip('''\
        # \[changelog\] entries

        New changelog entries in the `.yaml` file should follow a specific format and are only added for specified reasons (excising TOAs, adding/removing params, changing binary models, etc.). For more detailed instructions, run `new_changelog_entry?` in a new cell. This function can be used to format your entry, which should be added to the bottom of the appropriate `.yaml` file. Note: make sure your git `user.email` is properly configured, since this field is used to add your name to the entry.\
        ''',autorun)

    def write_out(self, filename="process.ipynb"):
        """
        Write out a new ipython notebook

        Parameters
        ==========
        filename (optional): notebook output filename, default "process.ipynb"
        """
        nb0 = new_notebook(cells=self.cells, metadata=self.metadata)
        f = codecs.open(filename, encoding='utf-8', mode='w')
        nbf.write(nb0, f, 4)
        f.close()



"""
This code contains the TimingConfiguration class, which is used to load configuration files and perform
actions, with items then passed to the notebooks.

Very basic usage:
    from timingconfiguration import TimingConfiguration
    tc = TimingConfiguration(CONFIGFILE)
"""
import io
import os
import re
import pint.toa as toa
import pint.models as model
import pint.fitter
import numpy as np
import astropy.units as u
from astropy import log
from astropy.time import Time
import yaml
import glob
from pint_pal.utils import write_if_changed, apply_cut_flag, apply_cut_select
from pint_pal.lite_utils import new_changelog_entry
from pint_pal.lite_utils import check_toa_version, check_tobs
import pint_pal.config

class TimingConfiguration:
    """
    This class contains the functionality to read
    from a configuration file and send that information
    to the timing notebooks.
    """
    def __init__(self, filename="config.yaml", tim_directory=None, par_directory=None):
        """
        Initialization method.

        Normally config files are written to be run from the root of a
        git checkout on the NANOGrav notebook server. If you want to run
        them from somewhere else, you may need to override these directories
        when you construct the TimingConfiguration object; this will not
        change what is recorded in the config file.

        Parameters
        ==========
        filename (optional) : path to the configuration file
        tim_directory (optional) : override the tim directory specified in the config
        par_directory (optional) : override the par directory specified in the config
        """
        self.filename = os.path.realpath(os.path.expanduser(filename))
        with open(self.filename) as FILE:
            self.config = yaml.load(FILE, Loader=yaml.FullLoader)
        if tim_directory is not None:
            self.config['tim-directory'] = tim_directory
        if par_directory is not None:
            self.config['par-directory'] = par_directory
        self.skip_check = self.config['skip-check'] if 'skip-check' in self.config.keys() else ''

    @property
    def tim_directory(self):
        """
        Location of tim files, as specified in the config.
        This returns the absolute path to the tim directory.
        """
        return os.path.realpath(
            os.path.join(pint_pal.config.DATA_ROOT, self.config['tim-directory'])
        )

    @tim_directory.setter
    def tim_directory(self, tim_directory):
        """
        Set tim directory.
        If a relative path is supplied, it will be turned into an absolute path.
        """
        self.config['tim-directory'] = tim_directory

    @property
    def par_directory(self):
        """
        Location of par files, as specified in the config.
        This returns the absolute path to the par directory.
        """
        return os.path.realpath(
            os.path.join(pint_pal.config.DATA_ROOT, self.config['par-directory'])
        )

    @par_directory.setter
    def par_directory(self, par_directory):
        """
        Set par directory.
        If a relative path is supplied, it will be turned into an absolute path.
        """
        self.config['par-directory'] = par_directory

    def get_source(self):
        """ Return the source name """
        return self.config['source']

    def get_model_path(self):
        """ Return full path to timing-model """
        if "timing-model" in self.config.keys() and self.config["timing-model"] is not None:
            return os.path.join(self.par_directory,self.config["timing-model"])
        return None

    def get_compare_model(self):
        """ Return the timing model file to compare with """
        if "compare-model" in self.config.keys() and self.config['compare-model'] is not None:
            return os.path.join(self.par_directory, self.config['compare-model'])
        return None

    def get_free_params(self, fitter):
        """Return list of free parameters"""
        if self.config["free-dmx"]:
            return self.config['free-params'] + [p for p in fitter.model.params if p.startswith("DMX_")]
        else:
            return self.config['free-params']
        
    #def get_febe_pairs(self):
    #    febe_list = []
    #    for f in to.orig_table['flags']:
    #        if f['pta'] == 'NANOGrav':
    #            febe_list.append(f['f'])
    #        elif f['pta'] == 'EPTA' or 'PPTA':
    #            febe_list.append(f['sys'])
    #        febe_pairs = set(f for f in febe_list)
    #    return febe_pairs


    def get_model_and_toas(self,usepickle=True,print_all_ignores=False,apply_initial_cuts=True,
            excised=False,pout_tim_path=None, include_pn=True):
        """Return the PINT model and TOA objects"""
        par_path = os.path.join(self.par_directory,self.config["timing-model"])
        toas = self.config["toas"]

        # configure settings appropriately if using excise.tim file
        if excised:
            if self.get_excised():
                toas = self.get_excised()
                self.tim_directory = ''
            else: # unset or file does not exist... 
                log.warning(f"excised-tim is unset or file does not exist")
                return None, None

        # capability to point to start from a pout.tim file
        if pout_tim_path:
            log.info(f"Loading TOAs from pout.tim file: {pout_tim_path}")
            toas = pout_tim_path
            self.tim_directory = ''
            if excised:
                log.warning(f"Set specific_tim_path=None (default) if excised-tim should be loaded.")
            excised = True # to avoid re-applying initial cuts

        # Individual tim file
        if isinstance(toas, str):
            toas = [toas]

        BIPM = self.get_bipm()
        EPHEM = self.get_ephem()
        m = model.get_model(par_path,allow_name_mixing=True)
        match = re.search(r"#\s+Created:\s+(\S+)", open(par_path).read())
        if match:
            m.meta['created_time'] = match.group(1)
            log.info(f"Par file created: {m.meta['created_time']}")
        m.meta['file_mtime'] = Time(os.path.getmtime(par_path), format="unix").isot


        if m.PSR.value != self.get_source():
            log.warning(f'{self.filename} source entry does not match par file value ({m.PSR.value}).')

        picklefilename = os.path.basename(self.filename) + ".pickle.gz"
        # Merge toa_objects (check this works for list of length 1)
        t = toa.get_TOAs([os.path.join(self.tim_directory,t) for t in toas],
                          usepickle=usepickle,
                          bipm_version=BIPM,
                          ephem=EPHEM,
                          planets=pint_pal.config.PLANET_SHAPIRO,
                          model=m,
                          picklefilename=picklefilename,
                          include_pn=include_pn
                        )

        # if we're dealing with wideband TOAs, each epoch has a single TOA, 
        # so don't bother checking to see if we can reduce entries
        #if self.get_toa_type() == "NB":
        #    self.check_for_bad_files(t, threshold=0.9, print_all=print_all_ignores)

        # Make a clean copy of original TOAs table (to track cut TOAs, flag_values)
        t.renumber(index_order=False)  # Renumber so the index column matches the order of TOAs
        assert np.all(t.table["index"]==np.arange(len(t)))
        t.orig_table = t.table.copy()

        #self.febe_set = set([f['f'] for f in t.orig_table['flags']])  # if this is necessary, set to None if f doesn't exist

        # If reading an intermediate (pout/excised) tim file, can simply apply cuts
        if excised:
            apply_initial_cuts = False
            apply_cut_select(t,reason='existing cuts, pre-excised')

        # Add 'cut' flags to TOAs according to config 'ignore' block.
        if apply_initial_cuts:
            self.backendset = set([f['be'] for f in t.orig_table['flags']])
            self.check_for_orphaned_recs(t)

            # Apply simul flags/cuts if appropriate
            if 'ASP' in self.backendset:
                self.check_simultaneous(t,'PUPPI','ASP')
            if 'GASP' in self.backendset:
                self.check_simultaneous(t,'GUPPI','GASP')

            t = self.apply_ignore(t,specify_keys=['orphaned-rec','mjd-start','mjd-end','bad-range',
                'snr-cut','poor-febe','orb-phase-range'],model=m)
            apply_cut_select(t,reason='initial cuts, specified keys')

        # Possibly better to handle this with toagen, but...ignore wb TOAs without pp_dm
        self.check_ppdm(t)

        # Issue with how PINT handles TOAs loaded from pickle vs. not (no pint version for the latter)
        if usepickle: check_toa_version(t)
        check_tobs(t)

        return m, t

    def get_summary(self):
        with io.StringIO() as o:
            # The two spaces at the end here ensure a line break (not a paragraph break) in Markdown
            print(f"Pulsar: {self.get_source()}  ", file=o)
            print(f"Mode: {self.get_toa_type()}  ", file=o)
            print(f"Par file: `{self.get_model_path()}`  ", file=o)
            print(f"Tim files:\n", file=o)
            toas = self.config["toas"]
            if isinstance(toas, str):
                toas = [toas]
            for t in toas:
                print(f"- `{t}`", file=o)
            print(file=o)
            return o.getvalue()

    def check_simultaneous(self,toas,backend1,backend2,warn=False):
        """Cut overlapped TOAs from the specified backends (simul)
    
        Assumes TOAs overlap if they are taken on the same day with the
        same receiver, should be fine for nanograv but could be made
        more picky to catch all cases. be1 is the backend to keep
        and be2 will be commented (e.g. PUPPI/ASP respectively). Also
        both toas will be marked with -simul flags.
    
        Parameters:
        ===========
        toas [pint.TOA]: PINT TOA object
        backend1 [string]: backend to keep if simultaneous (e.g. PUPPI)
        backend2 [string]: backend to cut if simultaneous (e.g. ASP)
        """
        cuts = np.array([f['cut'] if 'cut' in f else None for f in toas.orig_table['flags']])
        toaflags = toas.orig_table['flags']
        toamjds = toas.orig_table['mjd_float']
    
        idx1 = np.where([f['be']==backend1 for f in toas.orig_table['flags']])[0]
        idx2 = np.where([f['be']==backend2 for f in toas.orig_table['flags']])[0]
        simul_cut_inds = []
        for i1 in idx1:
            for i2 in idx2:
                if cuts[i2]: continue # Already cut
                if (toaflags[i1]['fe']==toaflags[i2]['fe']) and int(toamjds[i1])==int(toamjds[i2]):
                    if not freqs_overlap(toas.orig_table[i1],toas.orig_table[i2]): continue
                    if 'simul' not in toaflags[i1].keys(): # label and keep
                        toas.orig_table[i1]['flags']['simul'] = '1'
                    if 'simul' not in toaflags[i2].keys(): # label and cut
                        toas.orig_table[i2]['flags']['simul'] = '2'
                        simul_cut_inds.append(i2)
    
        if simul_cut_inds:
            apply_cut_flag(toas,np.array(simul_cut_inds),'simul',warn=warn)
            apply_cut_select(toas,f"simultaneous {backend1}/{backend2} observations")

    def check_file_outliers(self,toas,outpct_threshold=8.0):
        """ Check for files where Noutliers > nout_threshold, cut files where True (maxout)

        Parameters
        ==========
        toas: pint toas object
        outpct_threshold: float, optional
            cut file's remaining TOAs (maxout) if X% were flagged as outliers (default set by 5/64=8%)
        """
        names = np.array([f['name'] for f in toas.orig_table['flags']])
        cuts = np.array([f['cut'] if 'cut' in f else None for f in toas.orig_table['flags']])
        maxout_applied = False
        for name in set(names):
            nameinds = np.where([name == n for n in names])[0]
            ntoas_file = len(nameinds)
            nout_threshold = round(ntoas_file * outpct_threshold/100.0)
            file_cutlist = list(cuts[nameinds])
            outlier_cuts = ['outlier' in fc if fc else False for fc in file_cutlist]
            no_cuts = [not fc for fc in file_cutlist]
            if np.sum(outlier_cuts) > nout_threshold:
                if np.any(no_cuts):
                    log.warning(f"{name}: {outpct_threshold}% outlier threshold exceeded ({np.sum(outlier_cuts)}/{ntoas_file}), applying maxout cuts.")
                    dropinds = nameinds[np.array(no_cuts)]
                    apply_cut_flag(toas,dropinds,'maxout')
                    maxout_applied = True

        if maxout_applied: apply_cut_select(toas,reason=f"> {outpct_threshold}% outliers in file; maxout")

    def check_ppdm(self,toas):
        """ If wb type, check for pp_dm flags in all TOAs, else ignore """
        if self.get_toa_type() == 'WB':
            no_ppdm_inds = np.where(toas['pp_dm'] == '')[0]
            if no_ppdm_inds:
                log.warning(f"{len(no_ppdm_inds)} wb TOA(s) without pp_dm flag(s)?! Ignoring...")
                apply_cut_flag(toas,no_ppdm_inds,'no_ppdm')
                apply_cut_select(toas,reason='WB TOAs without pp_dm flags')

    def manual_cuts(self,toas,warn=False):
        """ Apply manual cuts after everything else and warn if redundant """
        toas = self.apply_ignore(toas,specify_keys=['bad-toa'],warn=warn)
        apply_cut_select(toas,reason='manual cuts, specified keys')
        
        toas = self.apply_ignore(toas,specify_keys=['bad-toa-averaged'],warn=warn)
        apply_cut_select(toas,reason='manual cuts, specified keys')

        toas = self.apply_ignore(toas,specify_keys=['bad-file'],warn=warn)
        apply_cut_select(toas,reason='manual cuts, specified keys')

    def count_bad_files(self):
        """ Return number of bad file entries """
        if "bad-file" in self.config['ignore'].keys():
            return len(self.config['ignore']['bad-file'])
        return None

    def count_bad_toas(self):
        """ Return number of bad toa entries """
        if "bad-toa" in self.config['ignore'].keys():
            return len(self.config['ignore']['bad-toa'])
        return None

    def get_bipm(self):
        """ Return the bipm string """
        if "bipm" in self.config.keys():
            return self.config['bipm']
        return None #return some default value instead?

    def get_ephem(self):
        """ Return the ephemeris string """
        if "ephem" in self.config.keys():
            return self.config['ephem']
        return None #return some default value instead?
    
    def get_febe_pairs(self,toas):
        febe_list = []
        for f in toas.orig_table['flags']:
            if f['pta'] == 'NANOGrav':
                febe_list.append(f['f'])
            elif f['pta'] == 'EPTA' or 'PPTA':
                febe_list.append(f['sys'])
            febe_pairs = set(f for f in febe_list)
        return febe_pairs

    def print_changelog(self):
        """Print changelog entries from .yaml in the notebook."""
        # If there's a changelog, write out its contents. If not, complain.
        if 'changelog' in self.config.keys():
            print('changelog:')
            if self.config['changelog'] is not None:
                for cl in self.config['changelog']:
                    print(f'  - {cl}')
            else:
                print('...no changelog entries currently exist.')
        else:
            print('YAML file does not include a changelog. Add \'changelog:\' and individual entries there.')

    def get_fitter(self):
        """ Return the fitter string (do more?) """
        if "fitter" in self.config.keys():
            return self.config['fitter']
        return None

    def construct_fitter(self, to, mo):
        """ Return the fitter, tracking pulse numbers if available """
        fitter_name = self.config['fitter']
        if fitter_name == 'Auto':
            return pint.fitter.Fitter.auto(to, mo)
        else:
            fitter_class = getattr(pint.fitter, fitter_name)
            return fitter_class(to, mo)

    def get_toa_type(self):
        """ Return the toa-type string """
        if "toa-type" in self.config.keys():
            return self.config['toa-type']
        return None

    def get_outfile_basename(self,ext=''):
        """ Return source.[nw]b basename (e.g. J1234+5678.nb) """
        basename = f'{self.get_source()}.{self.get_toa_type().lower()}'
        if ext: basename = '.'.join([basename,ext])
        return basename

    def get_niter(self):
        """ Return an integer of the number of iterations to fit """
        if "n-iterations" in self.config.keys():
            return int(self.config['n-iterations'])
        return 1

    def get_excised(self):
        """ Return excised-tim file if set and exists"""
        if 'excised-tim' in self.config['intermediate-results'].keys() and self.config['intermediate-results']['excised-tim']:
            if os.path.exists(self.config['intermediate-results']['excised-tim']):
                return self.config['intermediate-results']['excised-tim']
        return None

    def get_mjd_start(self):
        """Return mjd-start quantity (applies units days)"""
        if 'mjd-start' in self.config['ignore'].keys():
            return self.config['ignore']['mjd-start']
        return None

    def get_mjd_end(self):
        """Return mjd-end quantity (applies units days)"""
        if 'mjd-end' in self.config['ignore'].keys():
            return self.config['ignore']['mjd-end']
        return None

    def get_orphaned_rec(self):
        """Return orphaned receiver(s)"""
        if 'orphaned-rec' in self.config['ignore'].keys():
            return self.config['ignore']['orphaned-rec']
        return None 

    def get_bad_group(self):
        """Return bad group(s)"""
        if 'bad-group' in self.config['ignore'].keys():
            return self.config['ignore']['bad-group']
        return None
    
    def get_poor_febes(self):
        """Return poor frontend/backend combinations for removal"""
        if 'poor-febe' in self.config['ignore'].keys():
            return self.config['ignore']['poor-febe']
        return None

    def get_snr_cut(self):
        """ Return value of the TOA S/N cut """
        if "snr-cut" in self.config['ignore'].keys():
            return self.config['ignore']['snr-cut']
        return None #return some default value instead?
     
    def get_bad_files(self):
        """ Return list of bad files """
        if 'bad-file' in self.config['ignore'].keys():
            return self.config['ignore']['bad-file']
        return None
    
    def get_bad_ranges(self):
        """ Return list of bad file ranges by MJD ([MJD1,MJD2])"""
        if 'bad-range' in self.config['ignore'].keys():
            return self.config['ignore']['bad-range']
        return None

    def get_bad_toas(self):
        """ Return list of bad TOAs (lists: [filename, channel, subint]) """
        if 'bad-toa' in self.config['ignore'].keys():
            return self.config['ignore']['bad-toa']
        return None

    def get_bad_toas_averaged(self):
        """ Return list of bad TOAs (lists: [filename, channel, subint]) """
        if 'bad-toa-averaged' in self.config['ignore'].keys():
            return self.config['ignore']['bad-toa-averaged']
        return None

    
    def get_investigation_files(self):
        """ Makes a list from which the timer can choose which files they'd like to manually inspect"""
        ff_list = sorted(glob.glob('/nanograv/timing/releases/15y/toagen/data/*/*/*.ff'))
        match_files, match_toas = [], []
        # Note that you need the following check since this doesn't go through apply_ignore:
        if 'bad-file' in self.config['ignore'].keys() and self.config['ignore']['bad-file'] != None:
            for be in self.get_bad_files():
                if isinstance(be, list):
                    match_files.append([filenm for filenm in ff_list if be[0] in filenm])
                else: # bad-file entry is in the "old" style (just a string)
                    match_files.append([filenm for filenm in ff_list if be in filenm])
        if 'bad-toa' in self.config['ignore'].keys() and self.config['ignore']['bad-toa'] != None:
            for bt in self.get_bad_toas():
                match_toas.append([[filenm, bt[1], bt[2]] for filenm in ff_list if bt[0] in filenm])
        return sum(match_files,[]), sum(match_toas,[])
    
    def check_for_orphaned_recs(self, toas, nfiles_threshold=3):
        """Check for frontend/backend pairs that arise at or below threshold
        for number of files; also check that the set matches with those listed
        in the yaml.
        
        Parameters
        ==========
        toas: `pint.TOAs object`
        nfiles_threshold: int, optional
            Number of files at/below which a frontend/backend pair is orphaned.

        """
        #if toas.get_flag_value('f') != None
        #febe_pairs = set(toas.get_flag_value('f')[0])

        febe_pairs = set(toas.get_flag_value('f')[0])
        log.info(f'Frontend/backend pairs present in this data set: {febe_pairs}')


        febe_to_cut = []
        for febe in febe_pairs:
            f_bool = np.array([f == febe for f in toas.get_flag_value('f')[0]])
            f_names = toas[f_bool].get_flag_value('name')[0]
            files = set(f_names)
            n_files = len(files)
            if n_files > nfiles_threshold:
                log.info(f'{febe} files: {n_files}')
            else:
                febe_to_cut.append(febe)

        if febe_to_cut and ('orphaned-rec' in self.config['ignore'].keys()):
            ftc = set(febe_to_cut)
            if not self.get_orphaned_rec(): orph = set()
            else: orph = set(self.get_orphaned_rec())
            # Do sets of receivers to cut and those listed in the yaml match?
            if not (ftc == orph):
                # Add/remove from orphaned-rec?
                if (ftc - orph):
                    log.warning(f"{nfiles_threshold} or fewer files, add to orphaned-rec: {', '.join(ftc-orph)}")
                elif (orph - ftc):
                    log.warning(f"Remove from orphaned-rec: {', '.join(orph-ftc)}")
            else:
                pass     
        elif febe_to_cut:  # ...but no orphaned-rec field in the ignore block.
            febe_cut_str = ', '.join(febe_to_cut)
            log.warning(f"Add orphaned-rec to the ignore block in {self.filename}.")
            log.warning(f"{nfiles_threshold} or fewer files, add to orphaned-rec: {febe_cut_str}")
            print(f"Add the following line to {self.filename}...")
            new_changelog_entry('CURATE',f"orphaned receivers ({nfiles_threshold} or fewer files): {febe_cut_str}")

        return None 

    def check_outlier(self):
        """Perform simple checks on yaml outlier block and prob-outlier field
        """
        REQUIRED_KEYS = ['method','n-burn','n-samples']
        try:
            EXISTING_KEYS = self.config['outlier'].keys()
            VALUED_KEYS = [k for k in EXISTING_KEYS if self.config['outlier'][k] is not None]

            missing_required = set(REQUIRED_KEYS)-set(EXISTING_KEYS)
            if len(missing_required):
                log.warning(f'Required outlier keys not present: {missing_required}')

            invalid = set(EXISTING_KEYS) - set(REQUIRED_KEYS)
            if len(invalid):
                log.warning(f'Invalid outlier keys present: {invalid}')

            valid_null = set(EXISTING_KEYS) - set(VALUED_KEYS) - invalid
            if len(valid_null):
                log.warning(f'Required outlier keys included, but NOT in use: {valid_null}')

            # Does outlier block exist and are basic parameters set? Compare to OUTLIER_SAMPLES.
            valid_valued = set(VALUED_KEYS) - invalid
            if len(valid_valued) == len(REQUIRED_KEYS):
                log.info(f'Outlier analysis ({self.get_outlier_method()}) will run with {self.get_outlier_samples()} ({self.get_outlier_burn()} burn-in).')

        except KeyError:
            log.warning('outlier block should be added to your config file.')
            # print an example?

        # Does prob-outlier exist and is it set? Compare to OUTLIER_THRESHOLD.
        try:
            if self.get_prob_outlier():
                log.info(f'TOAs with outlier probabilities higher than {self.get_prob_outlier()} will be cut.')
            else:
                log.warning('The prob-outlier field in your ignore block must be set for outlier cuts to be made properly.')
        except KeyError:
            log.warning("prob-outlier field should be added to your config file's ignore block.")

    def check_for_bad_files(self, toas, threshold=0.9, print_all=False):
        """Check the bad-toas entries for files where more than a given
        percentange of TOAs have been flagged. Make appropriate suggestions
        for the user to update the `bad-file` entries, and optionally
        supply the revised `bad-toa` entries.

        Parameters
        ----------
        toas: pint.TOA
            A PINT TOA object that contains a table of TOAs loaded

        threshold: float
            A threshold fraction used to determine whether to suggest adding
            a bad-file line to the config file. Should be in the range [0, 1].
            Default is 0.9.

        print_all: bool
            If True, print both the suggested bad-file lines AND the revised
            bad-toa lines, where the new bad-toa lines now have entries from
            the suggested bad-files removed. Default is False.
        """
        # get the list of bad-toas already in the config file
        # only continue if that list has entries
        gotten_bad_toas = self.get_bad_toas()
        if isinstance(gotten_bad_toas, list):
            provided_bad_toas = [t[:3] for t in gotten_bad_toas] # ignores the 'reason' entry if present
            bad_toa_files = np.asarray(provided_bad_toas)[:, 0]
            # how many bad TOAs per file?
            unique, counts = np.unique(bad_toa_files, return_counts=True)
            bad_toa_file_counts = dict(zip(unique, counts))

            # how many raw TOAs per file?
            toa_files = toas.get_flag_value("name")[0]
            unique, counts = np.unique(toa_files, return_counts=True)
            toa_file_counts = dict(zip(unique, counts))

            # get the list of bad-files already in the config
            provided_bad_files = self.get_bad_files()
            if not isinstance(provided_bad_files, list):
                provided_bad_files = []

            # are there any files that have too many bad TOAs?
            new_bad_files = []
            for k in bad_toa_file_counts:
                # at this point, TOAs could have already been removed,
                # so check that the key exists first
                if k in toa_file_counts.keys():
                    n_toas = toa_file_counts[k]
                    n_bad = bad_toa_file_counts[k]
                    bad_frac = float(n_bad) / n_toas
                    # check that the bad fraction exceeds the threshold
                    # AND that the current file isn't already listed
                    if bad_frac >= threshold and k not in provided_bad_files:
                        new_bad_files.append(k)

            # only bother printing anything if there's a suggestion
            if len(new_bad_files) > 0:
                log.warn(
                    f"More than {threshold * 100}% of TOAs have been excised for some files"
                )
                log.info("Consider adding the following to `bad-file` in your config file:")
                for e in new_bad_files:
                    print(f"    - '{e}'")

            # if requested to update the bad-toa lines, figure out which
            # entries need to be removed
            if print_all:
                all_bad_files = np.concatenate((new_bad_files, provided_bad_files))
                bad_toas_to_del = []
                for e in all_bad_files:
                    _idx = np.where(bad_toa_files == e)[0]
                    bad_toas_to_del.extend(_idx)
                new_bad_toa_list = np.delete(np.asarray(provided_bad_toas), bad_toas_to_del, 0)
                log.info("The `bad-toa` list in your config file can be reduced to:")
                for t in new_bad_toa_list:
                    print(f"    - ['{t[0]}',{t[1]},{t[2]}]")

    def get_prob_outlier(self):
        if "prob-outlier" in self.config['ignore'].keys():
            return self.config['ignore']['prob-outlier']
        return None #return some default value instead?

    def get_noise_dir(self):
        """ Return base directory for noise results """
        if 'noise-dir' in self.config['intermediate-results'].keys():
            return self.config['intermediate-results']['noise-dir']
        return None
    
    def get_compare_noise_dir(self):
        """ Return base directory for noise results """
        if 'compare-noise-dir' in self.config['intermediate-results'].keys():
            return self.config['intermediate-results']['compare-noise-dir']
        return None

    def get_no_corner(self):
        """ Return boolean for no corner plot """
        if 'no-corner' in self.config['intermediate-results'].keys():
            return self.config['intermediate-results']['no-corner']
        return False

    def get_ignore_dmx(self):
        """ Return ignore-dmx toggle """
        if 'ignore-dmx' in self.config['dmx'].keys():
            return self.config['dmx']['ignore-dmx']
        return None

    def get_fratio(self):
        """ Return desired frequency ratio """
        if 'fratio' in self.config['dmx'].keys():
            return self.config['dmx']['fratio']
        return pint_pal.config.FREQUENCY_RATIO

    def get_sw_delay(self):
        """ Return desired max(solar wind delay) threshold """
        if 'max-sw-delay' in self.config['dmx'].keys():
            return self.config['dmx']['max-sw-delay']
        return pint_pal.config.MAX_SOLARWIND_DELAY

    def get_custom_dmx(self):
        """ Return MJD/binning params for handling DM events, etc. """
        if 'custom-dmx' in self.config['dmx'].keys():
            return self.config['dmx']['custom-dmx']
        return None

    def get_outlier_burn(self):
        """ Return outlier analysis burn-in samples """
        if 'n-burn' in self.config['outlier'].keys():
            return self.config['outlier']['n-burn']
        return None

    def get_outlier_samples(self):
        """ Return number of samples for outlier analysis """
        if 'n-samples' in self.config['outlier'].keys():
            return self.config['outlier']['n-samples']
        return None

    def get_outlier_method(self):
        """ Return outlier analysis method """
        if 'method' in self.config['outlier'].keys():
            return self.config['outlier']['method']
        return None

    def get_orb_phase_range(self):
        """ Return orb-phase-range to ignore """
        if 'orb-phase-range' in self.config['ignore'].keys():
            return self.config['ignore']['orb-phase-range']
        return None
    
    def get_check_cleared(self):
        """ Return bool to check if yaml has been cleared """
        if 'cleared' in self.config['check'].keys():
            return self.config['check']['cleared']
        return None

    def apply_ignore(self,toas,specify_keys=None,warn=False,model=None):
        """ Basic checks and return TOA excision info. """
        OPTIONAL_KEYS = ['mjd-start','mjd-end','snr-cut','bad-toa', 'bad-toa-averaged','bad-range','bad-file',
                        'orphaned-rec','bad-group','prob-outlier','poor-febe','orb-phase-range']
        EXISTING_KEYS = self.config['ignore'].keys()
        VALUED_KEYS = [k for k in EXISTING_KEYS if self.config['ignore'][k] is not None]

        missing_valid = set(OPTIONAL_KEYS)-set(EXISTING_KEYS)
        if len(missing_valid) and not specify_keys:
            log.info(f'Valid TOA excision keys not present: {missing_valid}')

        invalid = set(EXISTING_KEYS) - set(OPTIONAL_KEYS)
        if len(invalid):
            log.warning(f'Invalid TOA excision keys present: {invalid}')

        valid_null = set(EXISTING_KEYS) - set(VALUED_KEYS) - invalid
        if len(valid_null) and not specify_keys:
            log.info(f'TOA excision keys included, but NOT in use: {valid_null}')

        valid_valued = set(VALUED_KEYS) - invalid
        if len(valid_valued):
            # Provide capability to add -cut flags based on specific ignore fields
            if specify_keys is not None:
                valid_valued = valid_valued & set(specify_keys)
                log.info(f'Specified TOA excision keys: {valid_valued}')
            else:
                log.info(f'Valid TOA excision keys in use: {valid_valued}')
                 
        # All info here about selecting various TOAs.
        # Select TOAs to cut, then use apply_cut_flag.
        if 'orphaned-rec' in valid_valued:
            fs = np.array([(f['f'] if 'f' in f else None) for f in toas.orig_table['flags']])
            for o in self.get_orphaned_rec():
                orphinds = np.where(fs==o)[0]
                apply_cut_flag(toas,orphinds,'orphaned',warn=warn)
        if 'bad-group' in valid_valued:
            for flag, value in self.get_bad_group():
                vals = np.array([f[flag] for f in toas.orig_table['flags'] if flag in list(f.keys())])
                bad_inds = np.where(vals==value)[0]
                apply_cut_flag(toas,bad_inds,'badgroup',warn=warn)
        if 'mjd-start' in valid_valued:
            mjds = np.array([m for m in toas.orig_table['mjd_float']])
            startinds = np.where(mjds < self.get_mjd_start())[0]
            apply_cut_flag(toas,startinds,'mjdstart',warn=warn)
        if 'mjd-end' in valid_valued:
            mjds = np.array([m for m in toas.orig_table['mjd_float']])
            endinds = np.where(mjds > self.get_mjd_end())[0]
            apply_cut_flag(toas,endinds,'mjdend',warn=warn)
        if 'snr-cut' in valid_valued:
            if any('snr' not in f for f in toas.orig_table['flags']):
                log.warning("Excision key 'snr-cut' in use, but some TOAs are missing the 'snr' flag.")
            snrs = np.array([(float(f['snr']) if 'snr' in f else np.nan) for f in toas.orig_table['flags']])
            snrinds = np.where(snrs < self.get_snr_cut())[0]
            apply_cut_flag(toas,snrinds,'snr',warn=warn)
            if self.get_snr_cut() > 8.0 and self.get_toa_type() == 'NB':
                log.warning('snr-cut should be set to 8; try excising TOAs using other methods.')
            if self.get_snr_cut() > 25.0 and self.get_toa_type() == 'WB':
                log.warning('snr-cut should be set to 25; try excising TOAs using other methods.')
        if 'poor-febe' in valid_valued:
            fs = np.array([(f['f'] if 'f' in f else None) for f in toas.orig_table['flags']])
            for febe in self.get_poor_febes():
                febeinds = np.where(fs==febe)[0]
                apply_cut_flag(toas,febeinds,'poorfebe',warn=warn)
        if 'prob-outlier' in valid_valued:
            omethod = self.get_outlier_method().lower()  # accepts Gibbs and HMC, e.g.
            SUPPORTED_METHODS = ['gibbs','hmc']
            if omethod in SUPPORTED_METHODS:
                oflag = f'pout_{omethod}'
            else:
                log.warning(f'Outlier analysis method not recognized: {omethod}')
                oflag = f'pout_{omethod}'  # so that run doesn't crash
            pouts =  np.zeros(len(toas.orig_table))
            for i,fs in enumerate(toas.orig_table['flags']):
                if oflag in fs:
                    pouts[i] = float(fs[oflag])
            poutinds = np.where(pouts > self.get_prob_outlier())[0]
            oprob_flag = f'outlier{int(self.get_prob_outlier()*100)}'
            apply_cut_flag(toas,poutinds,oprob_flag,warn=warn)
        if 'bad-file' in valid_valued:
            logwarnfile = False
            names = np.array([f['name'] for f in toas.orig_table['flags']])
            cuts = np.array([f['cut'] if 'cut' in f else None for f in toas.orig_table['flags']])
            for be in self.get_bad_files():
                if isinstance(be, list): # either it's just a list, or a list with a reason
                    if len(be) == 1: # i.e. no reason given
                        logwarnfile = True
                    be = be[0]
                elif isinstance(be, str): # still in old format
                    logwarnfile = True

                fileinds = np.where([be in n for n in names])[0]
                # Check bad-file entry only matches one file
                name_matches = set(names[fileinds])
                if len(name_matches) > 1:
                    log.warning(f"Check {be} (matches multiple files): {name_matches}")
                    # Automatically explore matching files to see if any are immediately redundant.
                    for nm in name_matches:
                        matchinds = np.where([nm in n for n in names])[0]
                        remaining = np.array([not cut for cut in cuts[matchinds]])
                        alreadycut = np.invert(remaining)
                        if np.all(alreadycut):
                            log.warning(f"All TOAs from {nm} already cut: {set(cuts[matchinds][alreadycut])}")
                elif len(name_matches) == 1:
                    # Check bad-file entry is not redundant
                    remaining = np.array([not cut for cut in cuts[fileinds]])
                    alreadycut = np.invert(remaining)
                    if np.all(alreadycut):
                        log.warning(f"All TOAs from {be} already cut: {set(cuts[fileinds][alreadycut])}")
                    apply_cut_flag(toas,fileinds,'badfile',warn=warn)
                else:
                    log.warning(f"bad-file entry does not match any TOAs: {be}")

            if logwarnfile:
                log.warning(f'One or more bad-file entries lack reasons for excision; please add them.')
        if 'bad-range' in valid_valued:
            mjds = np.array([m for m in toas.orig_table['mjd_float']])
            backends = np.array([(f['be'] if 'be' in f else None) for f in toas.orig_table['flags']])
            for br in self.get_bad_ranges():
                if len(br) > 2:
                    rangeinds = np.where((mjds>br[0]) & (mjds<br[1]) & (backends==br[2]))[0]
                else:
                    rangeinds = np.where((mjds>br[0]) & (mjds<br[1]))[0]
                apply_cut_flag(toas,rangeinds,'badrange',warn=warn)
        if 'orb-phase-range' in valid_valued:
            mjds = np.array([m for m in toas.orig_table['mjd_float']])
            if not model:
                log.warning("Keyword argument 'model' is unset; cannot compute orbital phase, or apply eclipsing cut.")
            else:
                orbphase = model.orbital_phase(mjds, radians = False)
                phmin,phmax = self.get_orb_phase_range()
                phase_cut_inds = np.where((orbphase > phmin) & (orbphase < phmax))[0]
                apply_cut_flag(toas,phase_cut_inds,'eclipsing',warn=warn)

        if 'bad-toa' in valid_valued:
            logwarntoa = False
            names = np.array([f['name'] for f in toas.orig_table['flags']])
            subints = np.array([(int(f['subint']) if 'subint' in f else None) for f in toas.orig_table['flags']])
            if self.get_toa_type() == 'NB':
                chans = np.array([(int(f['chan']) if 'chan' in f else None) for f in toas.orig_table['flags']])
            btinds = []
            for bt in self.get_bad_toas():
                if len(bt) < 4: logwarntoa = True
                name,chan,subint = bt[:3]
                if self.get_toa_type() == 'NB':
                    bt_match = np.where((names==name) & (chans==chan) & (subints==subint))[0]
                else:
                    # don't match based on -chan flags, since WB TOAs don't have them
                    bt_match = np.where((names==name) & (subints==subint))[0]
                if len(bt_match): btinds.append(bt_match[0])
                else: log.warning(f"Listed bad TOA not matched: [{name}, {chan}, {subint}]")
            btinds = np.array(btinds)

            # Check for pre-existing cut flags if there are bad toas matched:
            cuts = np.array([f['cut'] if 'cut' in f else None for f in toas.orig_table['flags']])
            remaining = np.array([not cut for cut in cuts[btinds]])
            alreadycut = np.invert(remaining)

            if np.any(alreadycut):
                log.info(f"{np.sum(alreadycut)} bad-toa entries already cut: {set(cuts[btinds][alreadycut])}")
                log.info(f"bad-toa list can be reduced to {np.sum(remaining)} entries...")
                for i in btinds[remaining]:
                    if self.get_toa_type() == 'NB': print(f"  - [{names[i]},{chans[i]},{subints[i]}]")
                    else: print(f"  - [{names[i]},None,{subints[i]}]")

            if logwarntoa:
                log.warning(f'One or more bad-toa entries lack reasons for excision; please add them.')

            apply_cut_flag(toas,np.array(btinds),'badtoa',warn=warn)
        
        if 'bad-toa-averaged' in valid_valued:
            logwarntoa = False
            names = np.array([f['name'] for f in toas.orig_table['flags']])
            btinds = []
            for bt in self.get_bad_toas_averaged():
                name=bt[0]
                bt_match = np.where((names==name))[0]
                if len(bt_match): btinds.append(bt_match[0])
                else: log.warning(f"Listed bad TOA not matched: [{name}]")
            btinds = np.array(btinds)

            # Check for pre-existing cut flags if there are bad toas matched:
            cuts = np.array([f['cut'] if 'cut' in f else None for f in toas.orig_table['flags']])
            remaining = np.array([not cut for cut in cuts[btinds]])
            alreadycut = np.invert(remaining)

            if np.any(alreadycut):
                log.info(f"{np.sum(alreadycut)} bad-toa entries already cut: {set(cuts[btinds][alreadycut])}")
                log.info(f"bad-toa-averaged list can be reduced to {np.sum(remaining)} entries...")
                for i in btinds[remaining]:
                    print(f"  - [{names[i]},None,None]")

            if logwarntoa:
                log.warning(f'One or more bad-toa entries lack reasons for excision; please add them.')

            apply_cut_flag(toas,np.array(btinds),'badtoa-averaged',warn=warn)

        return toas

    def badtoa_index(self,badtoa,toas):
        """ Return index associated with bad-toa entry
        
        Parameters
        ==========
        badtoa: list, e.g. [filename,chan,subint,(reason)]
            Individual bad-toa entry; note that 'reason' is optional and not used;
            also wideband TOAs have chan=None.
        toas: `pint.TOAs object`
        """
        name,chan,subint = badtoa[:3]
        names = np.array([f['name'] for f in toas.orig_table['flags']])
        subints = np.array([int(f['subint']) for f in toas.orig_table['flags']])
        if self.get_toa_type() == 'NB':
            chans = np.array([int(f['chan']) for f in toas.orig_table['flags']])
            bt_match = np.where((names==name) & (chans==chan) & (subints==subint))[0]
        else:
            # don't match based on -chan flags, since WB TOAs don't have them
            bt_match = np.where((names==name) & (subints==subint))[0]
        if len(bt_match): return bt_match[0]
        return None

    def badtoa_info(self,badtoa,toas):
        """ Return notable information for bad-toa entry

        For formatting purposes, it is recommended this function is used in conjunction
        with others here, e.g. self.get_bad_toas(), though some guidance is provided
        below in case badtoa is entered manually.

        Parameters
        ==========
        badtoa: list, e.g. [filename,chan,subint,(reason)]
            Individual bad-toa entry; note that 'reason' is optional and not used;
            also wideband TOAs have chan=None.
        toas: `pint.TOAs object`
        """
        index = self.badtoa_index(badtoa,toas)
        toa = toas.orig_table[index]
        fe = toa['flags']['fe']
        good_fluxes = np.array([float(t['flags']['flux']) for t in toas.table]) # note: good toas only here
        fe_flags = np.array([t['flags']['fe'] for t in toas.table])
        fe_match = np.where(fe_flags==fe)[0]
        n_fe = len(fe_match)

        fe_fluxes = good_fluxes[fe_match]
        med_flux = np.median(fe_fluxes)
        # Might want a dictionary with useful values?

        # Some checks
        flo = np.percentile(fe_fluxes,5.)
        fhi = np.percentile(fe_fluxes,95.)
        extreme_flux = (float(toa['flags']['flux']) < flo) or (float(toa['flags']['flux']) > fhi)
        close_to_snrcut = (float(toa['flags']['snr']) - self.get_snr_cut()) < 1.0
        large_uncertainty = (toa['error'] > 50.0)  # microseconds

        # Should handle flux by frequency somehow in the check above and add something like:
        # Median flux value at XXX MHz is YYY based on ZZZ measurements (and possibly add percentiles)
        flux_err = f"{toa['flags']['flux']} +/- {toa['flags']['fluxe']}"
        index_line = f"    index: {index}"
        fe_line =    f"       fe: {fe}"
        error_line = f"    error: {toa['error']} us"
        snr_line =   f"      snr: {toa['flags']['snr']}"
        flux_line =  f"     flux: {flux_err} mJy (90% of {n_fe} {fe} TOAs in {flo:.2f}-{fhi:.2f} mJy range)"

        # Warnings added in a different color (bcolors.FAIL; see example below)
        # https://stackoverflow.com/questions/287871/how-to-print-colored-text-to-the-terminal
        if extreme_flux:
            flux_line = f"     flux: {flux_err} mJy"
            flux_line = f"{flux_line}  \033[91m ** far from median ({med_flux}) ** \033[0m"
        if close_to_snrcut:
            snr_line = f"{snr_line}  \033[91m ** close to snr-cut ({self.get_snr_cut()}) ** \033[0m"
        if large_uncertainty: 
            error_line = f"{error_line} \033[91m ** very large TOA uncertainty ** \033[0m"

        print(f"Info for bad-toa entry {badtoa}...")
        print(index_line)
        print(fe_line)
        print(error_line)
        print(snr_line)
        print(flux_line)
        print()

def freqs_overlap(toa1,toa2):
    """Returns true if TOAs from different backends overlap
    
    See check_simultaneous in TimingConfiguration
    
    Parameters:
    ===========
    toa1 [pint.TOA.table]: PINT TOA.table element
    toa2 [pint.TOA.table]: PINT TOA.table element
    """
    try: bw1 = float(toa1['flags']['bw'])
    except: bw1 = 0.0
    try: bw2 = float(toa2['flags']['bw'])
    except: bw2 = 0.0
    if toa1['freq']+bw1/2.0 < toa2['freq']-bw2/2.0: return False
    if toa2['freq']+bw2/2.0 < toa1['freq']-bw1/2.0: return False
    return True



'''
Unit testing for timingconfiguration.py

These tests are performed on YAML configuration files
'''


import unittest
from pint_pal.timingconfiguration import TimingConfiguration


class TimingConfigurationTests(unittest.TestCase):
    """ timingconfiguration.py testing class """

    def setUp(self):
        """ Load a TimingConfiguration object during setup """
        self.tc = TimingConfiguration("config/goodconfig.yaml")
        self.badtc = TimingConfiguration("configs/badconfig.yaml")


    def test_get_source(self):
        """ Check the reading of the source entry """
        self.assertEqual(self.tc.get_source(), "B1855+09")


    def test_get_model(self):
        """ Check the return of a PINT model object """
        self.assertEqual(self.tc.get_model().PSR.value, "B1855+09")
        with self.assertRaises(ValueError):
            self.badtc.get_model()


    def test_get_TOAs(self):
        """ Check the return of a PINT toa object, with various filters """
        pass


try:
    unittest.main(argv=[''], verbosity=2)
except SystemExit: #cleaner output below
    print



# Generic imports
import os, sys
import matplotlib.pyplot as plt
import numpy as np
from astropy import log
from multiprocessing import Pool

# Outlier/Epochalyptica imports
from pint.fitter import ConvergenceFailure
import copy
from scipy.special import fdtr
from pint_pal.utils import apply_cut_flag, apply_cut_select
from pint_pal.lite_utils import write_tim
from pint_pal.dmx_utils import *

def gibbs_run(entPintPulsar,results_dir=None,Nsamples=10000):
    """Necessary set-up to run gibbs sampler, and run it. Return pout.
    """
    # Imports
    import enterprise.signals.parameter as parameter
    from enterprise.signals import utils
    from enterprise.signals import signal_base
    from enterprise.signals.selections import Selection
    from enterprise.signals import white_signals
    from enterprise.signals import gp_signals
    from enterprise.signals.selections import Selection
    from enterprise.signals import selections
    from enterprise.signals import deterministic_signals
    from enterprise_outliers.gibbs_outlier import OutlierGibbs

    # white noise
    efac = parameter.Uniform(0.01,10.0)
    t2equad = parameter.Uniform(-10, -4)
    ecorr = parameter.Uniform(-10, -4)
    selection = selections.Selection(selections.by_backend)

    # white noise
    mn = white_signals.MeasurementNoise(efac=efac, log10_t2equad=t2equad, selection=selection)
    ec = gp_signals.EcorrBasisModel(log10_ecorr=ecorr, selection=selection)

    # red noise
    pl = utils.powerlaw(log10_A=parameter.Uniform(-18,-11),gamma=parameter.Uniform(0,7))
    rn = gp_signals.FourierBasisGP(spectrum=pl, components=30)

    # timing model
    tm = gp_signals.TimingModel()

    # combined signal
    s = mn + ec + rn + tm 

    # PTA
    pta = signal_base.PTA([s(entPintPulsar)])

    # Steve's code
    gibbs = OutlierGibbs(pta, model='mixture', vary_df=True,theta_prior='beta', vary_alpha=True)
    params = np.array([p.sample() for p in gibbs.params]).flatten()
    gibbs.sample(params, outdir=results_dir,niter=Nsamples, resume=False)
    poutlier = np.mean(gibbs.poutchain, axis = 0)

    #return np.mean(gibbs.poutchain, axis = 0)
    return poutlier

def get_entPintPulsar(model,toas,sort=False,drop_pintpsr=True):
    """Return enterprise.PintPulsar object

    Parameters
    ==========
    model: `pint.model.TimingModel` object
    toas: `pint.toa.TOAs` object
    sort: bool
        optional, default: False
    drop_pintpsr: bool
        optional, default: True; PintPulsar retains model/toas if False

    Returns
    =======
    model: `enterprise.PintPulsar` object
    """
    from enterprise.pulsar import PintPulsar
    return PintPulsar(toas,model,sort=sort,drop_pintpsr=drop_pintpsr)

def calculate_pout(model, toas, tc_object):
    """Determines TOA outlier probabilities using choices specified in the
    timing configuration file's outlier block. Write tim file with pout flags/values.

    Parameters
    ==========
    model: `pint.model.TimingModel` object
    toas: `pint.toa.TOAs` object
    tc_object: `pint_pal.timingconfiguration` object
    """
    method = tc_object.get_outlier_method()
    results_dir = f'outlier/{tc_object.get_outfile_basename()}'
    Nsamples = tc_object.get_outlier_samples()
    Nburnin = tc_object.get_outlier_burn()

    if method == 'hmc':
        epp = get_entPintPulsar(model, toas, drop_pintpsr=False)
        from enterprise_outliers.hmc_outlier import OutlierHMC
        pout = OutlierHMC(epp, outdir=results_dir, Nsamples=Nsamples, Nburnin=Nburnin)
        print('') # Progress bar doesn't print a newline
        # Some sorting will be needed here so pout refers to toas order?
    elif method == 'gibbs':
        epp = get_entPintPulsar(model, toas)
        pout = gibbs_run(epp,results_dir=results_dir,Nsamples=Nsamples)
    else:
        log.error(f'Specified method ({method}) is not recognized.')

    # Apply pout flags, cuts
    for i,oi in enumerate(toas.table['index']):
        toas.orig_table[oi]['flags'][f'pout_{method}'] = str(pout[i])

    # Re-introduce cut TOAs for writing tim file that includes -cut/-pout flags
    toas.table = toas.orig_table
    fo = tc_object.construct_fitter(toas,model)
    pout_timfile = f'{results_dir}/{tc_object.get_outfile_basename()}_pout.tim'
    write_tim(fo,toatype=tc_object.get_toa_type(),outfile=pout_timfile)

    # Need to mask TOAs once again
    apply_cut_select(toas,reason='resumption after write_tim, pout')

def make_pout_cuts(model,toas,tc_object,outpct_threshold=8.0):
    """Apply cut flags to TOAs with outlier probabilities larger than specified threshold.
    Also runs setup_dmx.

    Parameters
    ==========
    toas: `pint.toa.TOAs` object
    tc_object: `pint_pal.timingconfiguration` object
    outpct_threshold: float, optional
       cut file's remaining TOAs (maxout) if X% were flagged as outliers (default set by 5/64=8%) 
    """
    toas = tc_object.apply_ignore(toas,specify_keys=['prob-outlier'])
    apply_cut_select(toas,reason='outlier analysis, specified key')
    toas = setup_dmx(model,toas,frequency_ratio=tc_object.get_fratio(),max_delta_t=tc_object.get_sw_delay())

    # Now cut files if X% or more TOAs/file are flagged as outliers
    if tc_object.get_toa_type() == 'NB':
        tc_object.check_file_outliers(toas,outpct_threshold=outpct_threshold)
        toas = setup_dmx(model,toas,frequency_ratio=tc_object.get_fratio(),max_delta_t=tc_object.get_sw_delay())
    else:
        log.info('Skipping maxout cuts (wideband).')

def Ftest(chi2_1, dof_1, chi2_2, dof_2):
    """
    Ftest(chi2_1, dof_1, chi2_2, dof_2):
        Compute an F-test to see if a model with extra parameters is
        significant compared to a simpler model.  The input values are the
        (non-reduced) chi^2 values and the numbers of DOF for '1' the
        original model and '2' for the new model (with more fit params).
        The probability is computed exactly like Sherpa's F-test routine
        (in Ciao) and is also described in the Wikipedia article on the
        F-test:  http://en.wikipedia.org/wiki/F-test
        The returned value is the probability that the improvement in
        chi2 is due to chance (i.e. a low probability means that the
        new fit is quantitatively better, while a value near 1 means
        that the new model should likely be rejected).
        If the new model has a higher chi^2 than the original model,
        returns value of False
    """
    delta_chi2 = chi2_1 - chi2_2
    if delta_chi2 > 0:
      delta_dof = dof_1 - dof_2
      new_redchi2 = chi2_2 / dof_2
      F = (delta_chi2 / delta_dof) / new_redchi2
      ft = 1.0 - fdtr(delta_dof, dof_2, F)
    else:
      ft = False
    return ft

# This global var allows the (unpickleable) PINT model object
# to be passed to the multiprocessing workers in epochalyptica.
_epoch_args = None

def _set_epoch_args(model, toas, tc_object):
    """Sets arguments for test_one_epoch() into a global variable 
    for use in multiprocessing."""
    global _epoch_args
    _epoch_args = (model, toas, tc_object)

def _test_one_epoch_args(filename):
    """Single-argument wrapper function for test_one_epoch() for use with
    multiprocessing."""
    return test_one_epoch(*_epoch_args, filename)

def test_one_epoch(model, toas, tc_object, filename):
    """Test chi2 for removal of one epoch (filename).  Used internally
    by epochalyptica().

    Returns:
      receiver - receiver name of the removed file
      mjd - MJD of the removed file
      chi2 - post-fit chi2 after removing the file
      ndof - post-fit NDOF after removing the file
      ntoas - number of TOAs remaining after removal
      esum - weighted sum of removed TOA uncertainties
    """
    using_wideband = tc_object.get_toa_type() == 'WB'
    log.info(f"Testing removal of {filename} ntoas={toas.ntoas}")

    maskarray = np.ones(toas.ntoas,dtype=bool)
    receiver = None
    mjd = None
    toaval = None
    dmxindex = None
    dmxlower = None
    dmxupper = None
    esum = 0.0
    # Note, t[1]: mjd, t[2]: mjd (d), t[3]: error (us), t[6]: flags dict
    for index,t in enumerate(toas.table):
        if t[6]['name'] == filename:
            if receiver == None:
                receiver = t[6]['f']
            if mjd == None:
                mjd = int(t[1].value)
            if toaval == None:
                toaval = t[2]
                i = 1
                while dmxindex == None:
                    DMXval = f"DMXR1_{i:04d}"
                    lowerbound = getattr(model.components['DispersionDMX'],DMXval).value
                    DMXval = f"DMXR2_{i:04d}"
                    upperbound = getattr(model.components['DispersionDMX'],DMXval).value
                    if toaval > lowerbound and toaval < upperbound:
                        dmxindex = f"{i:04d}"
                        dmxlower = lowerbound
                        dmxupper = upperbound
                    i += 1
            esum = esum + 1.0 / (float(t[3])**2.0)
            maskarray[index] = False

    toas.select(maskarray)
    numtoas_in_dmxrange = 0
    for toa in toas.table:
        if toa[2] > dmxlower and toa[2] < dmxupper:
            numtoas_in_dmxrange += 1
    newmodel = model
    if numtoas_in_dmxrange == 0:
        log.debug(f"Removing DMX range {dmxindex}")
        newmodel = copy.deepcopy(model)
        newmodel.components['DispersionDMX'].remove_param(f'DMXR1_{dmxindex}')
        newmodel.components['DispersionDMX'].remove_param(f'DMXR2_{dmxindex}')
        newmodel.components['DispersionDMX'].remove_param(f'DMX_{dmxindex}')
    f = tc_object.construct_fitter(toas,newmodel)
    try:
        f.fit_toas(maxiter=tc_object.get_niter())
    except ConvergenceFailure:
        log.info('Failed to converge; moving on with best result.')
    ndof, chi2 = f.resids.dof, f.resids.chi2
    ntoas = toas.ntoas
    esum = 1.0 / np.sqrt(esum)
    toas.unselect()
    return receiver, mjd, chi2, ndof, ntoas, esum

def epochalyptica(model,toas,tc_object,ftest_threshold=1.0e-6,nproc=1):
    """ Test for the presence of remaining bad epochs (files) by removing one at a
        time and examining its impact on the residuals; pre/post reduced
        chi-squared values are assessed using an F-statistic.  

    Parameters:
    ===========
    model: `pint.model.TimingModel` object
    toas: `pint.toa.TOAs` object
    tc_object: `pint_pal.timingconfiguration` object
    ftest_threshold: float
        optional, threshold below which files will be dropped
    nproc: number of parallel processes to use for tests
    """
    using_wideband = tc_object.get_toa_type() == 'WB'
    f_init = tc_object.construct_fitter(toas,model)
    try:
        f_init.fit_toas(maxiter=tc_object.get_niter())
    except ConvergenceFailure:
        log.info('Failed to converge; moving on with best result.')
    ndof_init, chi2_init = f_init.resids.dof, f_init.resids.chi2
    ntoas_init = toas.ntoas  # How does this change for wb?
    redchi2_init = chi2_init / ndof_init

    filenames = sorted(set(toas.get_flag_value('name')[0]))
    outdir = f'outlier/{tc_object.get_outfile_basename()}'
    outfile = os.path.join(outdir,'epochdrop.txt')

    # Check for existence of path and make directories if they don't exist
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    fout = open(outfile,'w')
    numfiles = len(filenames)
    log.info(f'There are {numfiles} files to analyze.')
    files_to_drop = []

    # Run tests in parallel
    results = None
    if nproc > 1:
        with Pool(nproc, _set_epoch_args, (f_init.model, toas, tc_object)) as p:
            results = p.map(_test_one_epoch_args, filenames)

    for i, filename in enumerate(filenames):
        if results is not None:
            receiver, mjd, chi2, ndof, ntoas, esum = results[i]
        else: 
            receiver, mjd, chi2, ndof, ntoas, esum = test_one_epoch(f_init.model, toas, tc_object, filename)
        redchi2 = chi2 / ndof
        log.debug(f"After masking TOA(s) from {filename}...")
        log.debug(f"ndof init: {ndof_init}, ndof trial: {ndof}; chi2 init: {chi2_init}, chi2 trial: {chi2}")
        if ndof_init != ndof:
            ftest = Ftest(float(chi2_init),int(ndof_init),float(chi2),int(ndof))
            if ftest < ftest_threshold: files_to_drop.append(filename)
            log.debug(f"ftest: {ftest}")
        else:
            ftest = False
        fout.write(f"{filename} {receiver} {mjd:d} {(ntoas_init - ntoas):d} {ftest:e} {esum}\n")
        fout.flush()
    fout.close()

    # Apply cut flags
    names = np.array([f['name'] for f in toas.orig_table['flags']])
    for ftd in files_to_drop:
        filedropinds = np.where(names==ftd)[0]
        apply_cut_flag(toas,filedropinds,'epochdrop')

    # Make cuts, fix DMX windows if necessary
    if len(files_to_drop):
        apply_cut_select(toas,reason='epoch drop analysis')
        toas = setup_dmx(model,toas,frequency_ratio=tc_object.get_fratio(),max_delta_t=tc_object.get_sw_delay())
    else:
        log.info('No files dropped (epochalyptica).')

    # Re-introduce cut TOAs for writing tim file that includes -cut flags
    toas.table = toas.orig_table
    fo = tc_object.construct_fitter(toas,model)
    excise_timfile = f'{outdir}/{tc_object.get_outfile_basename()}_excise.tim'
    write_tim(fo,toatype=tc_object.get_toa_type(),outfile=excise_timfile)

    # Need to mask TOAs once again
    apply_cut_select(toas,reason='resumption after write_tim (excise)')
